# Copyright (C) 2023-2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import itertools
import logging
import os
import shutil

import attr
import pytest
import yaml

from swh.model.model import Content, Origin
from swh.model.swhids import ExtendedSWHID

from ..recovery_bundle import (
    ContentDataNotFound,
    Manifest,
    RecoveryBundle,
    RecoveryBundleCreator,
    SecretRecoveryError,
    SecretSharing,
    UnsupportedFeatureException,
    age_decrypt,
    age_encrypt,
    convert_bits,
    generate_age_keypair,
    list_yubikey_identities,
    recover_object_decryption_key_from_encrypted_shares,
)
from .conftest import (
    ALI_PUBLIC_KEY,
    ALI_SECRET_KEY,
    BOB_PUBLIC_KEY,
    BOB_SECRET_KEY,
    CAMILLE_PUBLIC_KEY,
    CAMILLE_SECRET_KEY,
    DLIQUE_PUBLIC_KEY,
    DLIQUE_SECRET_KEY,
    ESSUN_PUBLIC_KEY,
    ESSUN_SECRET_KEY,
    OBJECT_PUBLIC_KEY,
    OBJECT_SECRET_KEY,
    TWO_GROUPS_REQUIRED_WITH_ONE_MINIMUM_SHARE_EACH_SECRET_SHARING_YAML,
    object_decryption_key_provider_for_sample,
)


@pytest.fixture
def manifest_dict():
    return {
        "version": 3,
        "removal_identifier": "TDN-2023-06-18-01",
        "created": datetime.datetime(
            2023, 6, 18, 13, 12, 42, tzinfo=datetime.timezone.utc
        ),
        "requested": [
            Origin(url="https://example.com/swh/graph2"),
            ExtendedSWHID.from_string(
                "swh:1:snp:0000000000000000000000000000000000000022"
            ),
        ],
        "swhids": [
            ExtendedSWHID.from_string(s)
            for s in (
                "swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165",
                "swh:1:snp:0000000000000000000000000000000000000022",
                "swh:1:rel:0000000000000000000000000000000000000021",
                "swh:1:rev:0000000000000000000000000000000000000018",
                "swh:1:rev:0000000000000000000000000000000000000013",
                "swh:1:dir:0000000000000000000000000000000000000017",
                "swh:1:cnt:0000000000000000000000000000000000000016",
                "swh:1:cnt:0000000000000000000000000000000000000012",
                "swh:1:cnt:0000000000000000000000000000000000000015",
                "swh:1:cnt:0000000000000000000000000000000000000014",
                "swh:1:cnt:0000000000000000000000000000000000000011",
            )
        ],
        "referencing": [
            ExtendedSWHID.from_string(s)
            for s in (
                "swh:1:rel:0000000000000000000000000000000000000010",
                "swh:1:rev:0000000000000000000000000000000000000009",
                "swh:1:dir:0000000000000000000000000000000000000012",
            )
        ],
        "decryption_key_shares": {
            "YubiKey serial 4245067 slot 1": "-----BEGIN AGE ENCRYPTED FILE-----\n"
            "YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IHBpdi1wMjU2IHcvb0k0USBBb3FMYjRM\n"
            "V3dlcm9YazZkTU9UZld4eEVhYUlBZHRBQ05CQndOUFZJMmV1NApmNTY1MUJFdks1\n"
            "aE9TZzQ3NFJGN0cvQlFIMDZNSTkxUEpOblJteUkyK2FVCi0+IDxYTSFKLWdyZWFz\n"
            "ZSBCfWErZHkKNEMrbTdqekhTZTQ4c3pXRGZjK3N0UTh2Qi9ISU1XdFF6a0RvdmRl\n"
            "NAotLS0gYk9Ob2dkUTJRZE9nT3BTK29JWU5pRkZIVC9pUzJQaHRZc05sMjd6S1Rr\n"
            "OAoRXkzBiNX98H+353sOjGxJvCdYmtUdn7ozR35g+VSB6zxS972s2drkuKxQ0kIN\n"
            "MIjaytf/RJ0J3N/x8CtsEvXSoGjnuIT0GuEUbCqG0Qg0/YrrDzEGcD34l6JnD187\n"
            "5nVFnUimLXK6S2HeEDTJUZuLWfmglqaZaZjPnEKxqu8TfrJDBgg7miJLC+rGXhn9\n"
            "4ArtFIaOQgotCHZ8Y0lpmqGJIVTKWgdgpW+JjzyG\n"
            "-----END AGE ENCRYPTED FILE-----\n",
            "YubiKey serial 5229836 slot 1": "-----BEGIN AGE ENCRYPTED FILE-----\n"
            "YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IHBpdi1wMjU2IHcvb0k0USBBaTZhaUo3\n"
            "WnMzMmlTUlp5QmNhTkI1bHlmcHNyY0FPQ0RnK1BQdHQxS0EvbAppVnExb3BZcFRW\n"
            "ZkZ1ZFZrQWlyaU9HTkRKREYvU2tSaldkSHpWdVd1aGFVCi0+IDUrPVssLWdyZWFz\n"
            "ZQpzcm1WSkNqOWVrOU5GUXRMSmpFVVR4aEhrM0UKLS0tIFl2QkN6d1QzdWN6U0dB\n"
            "VHVzYk1SdDBLNlhNanJGc2x4L2hMZTZrSUxTSGMKLOKIpGZtKtUeOsSrcoIvKiBu\n"
            "DAoLXMGY+302lQRJsdJ3I7N+eFhRATsOM7vO8eupXbee87kIkGB7GaqGR5X48GR1\n"
            "oNrMsY5PcjZICxLjWYX9cMVMAXcmBjV9ZCWwqzmw86rY0k74mRwhE0dYd95P90+5\n"
            "NniuNgxQYKkM5QoKVHn36ISJGUgcvp5/JCM69X7kM8UvjLarFeYdHfqqAZUImNla\n"
            "lEdIqdOmnUs=\n"
            "-----END AGE ENCRYPTED FILE-----\n",
        },
        "reason": "copyright issue",
        "expire": datetime.datetime(
            2024, 6, 18, 13, 12, 42, tzinfo=datetime.timezone.utc
        ),
    }


@pytest.fixture
def manifest_dict_dumpable(manifest_dict):
    manifest_dict["requested"] = [
        x.url if isinstance(x, Origin) else str(x) for x in manifest_dict["requested"]
    ]
    manifest_dict["swhids"] = [str(s) for s in manifest_dict["swhids"]]
    manifest_dict["referencing"] = [str(s) for s in manifest_dict["referencing"]]
    return manifest_dict


def test_manifest_load_success(manifest_dict_dumpable):
    assert Manifest.load(yaml.dump(manifest_dict_dumpable))


def test_manifest_load_success_with_no_optionals(manifest_dict_dumpable):
    del manifest_dict_dumpable["reason"]
    del manifest_dict_dumpable["expire"]
    assert Manifest.load(yaml.dump(manifest_dict_dumpable))


@pytest.mark.parametrize(
    "invalid_manifest_dict",
    [
        pytest.param({"version": 42}, id="invalid_version"),
        pytest.param({"expire": "2024-06-18T13:12:42Z"}, id="str_instead_of_datetime"),
        pytest.param({"requested": []}, id="empty_requested"),
        pytest.param({"swhids": []}, id="empty_swhids"),
        pytest.param({"decryption_key_shares": {}}, id="empty_shares"),
        pytest.param({"invalid": "field"}, id="invalid_field"),
    ],
)
def test_manifest_load_failure(manifest_dict_dumpable, invalid_manifest_dict):
    manifest_dict_dumpable.update(invalid_manifest_dict)
    with pytest.raises((ValueError, TypeError)):
        assert Manifest.load(yaml.dump(manifest_dict_dumpable))


EXPECTED_MANIFEST_DUMP = """\
version: 3
removal_identifier: TDN-2023-06-18-01
created: 2023-06-18T13:12:42+00:00
requested:
- https://example.com/swh/graph2
- swh:1:snp:0000000000000000000000000000000000000022
swhids:
- swh:1:ori:8f50d3f60eae370ddbf85c86219c55108a350165
- swh:1:snp:0000000000000000000000000000000000000022
- swh:1:rel:0000000000000000000000000000000000000021
- swh:1:rev:0000000000000000000000000000000000000018
- swh:1:rev:0000000000000000000000000000000000000013
- swh:1:dir:0000000000000000000000000000000000000017
- swh:1:cnt:0000000000000000000000000000000000000016
- swh:1:cnt:0000000000000000000000000000000000000012
- swh:1:cnt:0000000000000000000000000000000000000015
- swh:1:cnt:0000000000000000000000000000000000000014
- swh:1:cnt:0000000000000000000000000000000000000011
referencing:
- swh:1:rel:0000000000000000000000000000000000000010
- swh:1:rev:0000000000000000000000000000000000000009
- swh:1:dir:0000000000000000000000000000000000000012
decryption_key_shares:
  YubiKey serial 4245067 slot 1: |
    -----BEGIN AGE ENCRYPTED FILE-----
    YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IHBpdi1wMjU2IHcvb0k0USBBb3FMYjRM
    V3dlcm9YazZkTU9UZld4eEVhYUlBZHRBQ05CQndOUFZJMmV1NApmNTY1MUJFdks1
    aE9TZzQ3NFJGN0cvQlFIMDZNSTkxUEpOblJteUkyK2FVCi0+IDxYTSFKLWdyZWFz
    ZSBCfWErZHkKNEMrbTdqekhTZTQ4c3pXRGZjK3N0UTh2Qi9ISU1XdFF6a0RvdmRl
    NAotLS0gYk9Ob2dkUTJRZE9nT3BTK29JWU5pRkZIVC9pUzJQaHRZc05sMjd6S1Rr
    OAoRXkzBiNX98H+353sOjGxJvCdYmtUdn7ozR35g+VSB6zxS972s2drkuKxQ0kIN
    MIjaytf/RJ0J3N/x8CtsEvXSoGjnuIT0GuEUbCqG0Qg0/YrrDzEGcD34l6JnD187
    5nVFnUimLXK6S2HeEDTJUZuLWfmglqaZaZjPnEKxqu8TfrJDBgg7miJLC+rGXhn9
    4ArtFIaOQgotCHZ8Y0lpmqGJIVTKWgdgpW+JjzyG
    -----END AGE ENCRYPTED FILE-----
  YubiKey serial 5229836 slot 1: |
    -----BEGIN AGE ENCRYPTED FILE-----
    YWdlLWVuY3J5cHRpb24ub3JnL3YxCi0+IHBpdi1wMjU2IHcvb0k0USBBaTZhaUo3
    WnMzMmlTUlp5QmNhTkI1bHlmcHNyY0FPQ0RnK1BQdHQxS0EvbAppVnExb3BZcFRW
    ZkZ1ZFZrQWlyaU9HTkRKREYvU2tSaldkSHpWdVd1aGFVCi0+IDUrPVssLWdyZWFz
    ZQpzcm1WSkNqOWVrOU5GUXRMSmpFVVR4aEhrM0UKLS0tIFl2QkN6d1QzdWN6U0dB
    VHVzYk1SdDBLNlhNanJGc2x4L2hMZTZrSUxTSGMKLOKIpGZtKtUeOsSrcoIvKiBu
    DAoLXMGY+302lQRJsdJ3I7N+eFhRATsOM7vO8eupXbee87kIkGB7GaqGR5X48GR1
    oNrMsY5PcjZICxLjWYX9cMVMAXcmBjV9ZCWwqzmw86rY0k74mRwhE0dYd95P90+5
    NniuNgxQYKkM5QoKVHn36ISJGUgcvp5/JCM69X7kM8UvjLarFeYdHfqqAZUImNla
    lEdIqdOmnUs=
    -----END AGE ENCRYPTED FILE-----
reason: copyright issue
expire: 2024-06-18T13:12:42+00:00
"""


def test_manifest_dump(manifest_dict):
    manifest = Manifest(**manifest_dict)
    assert manifest.dump() == EXPECTED_MANIFEST_DUMP


def test_manifest_dump_no_optionals(manifest_dict):
    del manifest_dict["reason"]
    del manifest_dict["expire"]
    manifest = Manifest(**manifest_dict)
    dump = manifest.dump()
    assert "reason:" not in dump
    assert "expire:" not in dump


def test_manifest_load():
    manifest = Manifest.load(EXPECTED_MANIFEST_DUMP)
    assert isinstance(manifest, Manifest)


def test_manifest_dump_load_roundtrip(manifest_dict):
    manifest = Manifest(**manifest_dict)
    assert manifest == Manifest.load(manifest.dump())


def test_generate_age_keypair():
    public_key, secret_key = generate_age_keypair()
    assert public_key.startswith("age")
    assert len(public_key) == 62
    assert secret_key.startswith("AGE-SECRET-KEY-")
    assert len(secret_key) == 74


def test_age_encrypt_decrypt_roundtrip():
    public_key = OBJECT_PUBLIC_KEY
    secret_key = OBJECT_SECRET_KEY
    cleartext = b"All Your Source Code Base Must Belong To Us!"
    ciphertext = age_encrypt(public_key, cleartext)
    assert age_decrypt(secret_key, ciphertext) == cleartext


def test_convert_5bits_to_8bits():
    assert [104, 101, 108, 108, 111] == convert_bits(
        [13, 1, 18, 22, 24, 27, 3, 15], 5, 8
    )


def test_convert_8bits_to_5bits():
    assert convert_bits([104, 101, 108, 108, 111], 8, 5) == [
        13,
        1,
        18,
        22,
        24,
        27,
        3,
        15,
    ]


EXAMPLE_SECRET_SHARING_YAML = """\
secret_sharing:
  minimum_required_groups: 2
  groups:
    legal:
      minimum_required_shares: 1
      recipient_keys:
        "YubiKey serial 4245067 slot 1": |-
          age1yubikey1q2e37f74zzazz75mtggzql3at66pegemfnul0dtd7axctahljkvsqezscaq
        "YubiKey serial 2284622 slot 3": |-
          age1yubikey1q0ucnwg558zcwrc752evk3620q2t4mkwz6a0lq9u3clsfmealsmlz330kz2
    sysadmins:
      minimum_required_shares: 1
      recipient_keys:
        "YubiKey serial 3862152 slot 1": |-
          age1yubikey1q04k5fs8cz6kypt7vjetl2gc2qtpz9lyzpxrvv2agzt6h8n3awmzk9sgd8v
        "YubiKey serial 6927448 slot 1": |-
          age1yubikey1qt2p377vq6qg58l8gaframp9yggvsysddraa72aehma5mw623r8rqk0mlgu
"""


@pytest.fixture
def example_secret_sharing():
    return SecretSharing.from_dict(
        yaml.safe_load(EXAMPLE_SECRET_SHARING_YAML)["secret_sharing"]
    )


@pytest.mark.skipif(
    not shutil.which("age-plugin-yubikey"),
    reason="missing `age-plugin-yubikey` executable",
)
def test_generate_encrypted_shares(example_secret_sharing):
    encrypted_shares = example_secret_sharing.generate_encrypted_shares(
        "example", OBJECT_SECRET_KEY
    )
    assert len(encrypted_shares) == 4
    assert encrypted_shares.keys() == {
        "YubiKey serial 4245067 slot 1",
        "YubiKey serial 2284622 slot 3",
        "YubiKey serial 3862152 slot 1",
        "YubiKey serial 6927448 slot 1",
    }
    assert all(
        s.startswith("-----BEGIN AGE ENCRYPTED FILE-----")
        for s in encrypted_shares.values()
    )


@pytest.fixture
def secret_sharing_2_groups_required_with_1_minimum_each():
    return SecretSharing.from_dict(
        yaml.safe_load(
            TWO_GROUPS_REQUIRED_WITH_ONE_MINIMUM_SHARE_EACH_SECRET_SHARING_YAML
        )["secret_sharing"]
    )


def available_secret_keys_for_2_groups_required_with_1_minimum_each():
    yield ("Dlique", DLIQUE_SECRET_KEY)
    yield ("Ali", ALI_SECRET_KEY)


TWO_GROUPS_REQUIRED_OF_THREE_WITH_ONE_AND_TWO_MINIMUM_IN_EACH_SECRET_SHARING_YAML = f"""\
secret_sharing:
  minimum_required_groups: 2
  groups:
    legal:
      minimum_required_shares: 1
      recipient_keys:
        "Ali": {ALI_PUBLIC_KEY}
    dpo:
      minimum_required_shares: 1
      recipient_keys:
        "Bob": {BOB_PUBLIC_KEY}
    sysadmins:
      minimum_required_shares: 2
      recipient_keys:
        "Camille": {CAMILLE_PUBLIC_KEY}
        "Dlique": {DLIQUE_PUBLIC_KEY}
        "Essun": {ESSUN_PUBLIC_KEY}
"""


@pytest.fixture
def secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each():
    return SecretSharing.from_dict(
        yaml.safe_load(
            TWO_GROUPS_REQUIRED_OF_THREE_WITH_ONE_AND_TWO_MINIMUM_IN_EACH_SECRET_SHARING_YAML
        )["secret_sharing"]
    )


def available_secret_keys_for_2_groups_required_of_3_only_ones():
    yield ("Ali", ALI_SECRET_KEY)
    yield ("Bob", BOB_SECRET_KEY)


def available_secret_keys_for_2_groups_required_of_3_one_and_two():
    yield ("Essun", ESSUN_SECRET_KEY)
    yield ("Bob", BOB_SECRET_KEY)
    yield ("Camille", CAMILLE_SECRET_KEY)


@pytest.mark.parametrize(
    "secret_sharing, available_secret_keys",
    [
        (
            "secret_sharing_2_groups_required_with_1_minimum_each",
            available_secret_keys_for_2_groups_required_with_1_minimum_each,
        ),
        (
            "secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each",
            available_secret_keys_for_2_groups_required_of_3_only_ones,
        ),
        (
            "secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each",
            available_secret_keys_for_2_groups_required_of_3_one_and_two,
        ),
    ],
)
def test_object_decryption_key_recovery_roundtrip(
    request, secret_sharing, available_secret_keys
):
    secret_sharing = request.getfixturevalue(secret_sharing)

    encrypted_shares = secret_sharing.generate_encrypted_shares(
        secret_sharing, OBJECT_SECRET_KEY
    )
    recovered_key = recover_object_decryption_key_from_encrypted_shares(
        encrypted_shares, available_secret_keys
    )
    assert recovered_key == OBJECT_SECRET_KEY


def test_object_decryption_key_recovery_with_not_enough_secret_keys(
    secret_sharing_2_groups_required_with_1_minimum_each,
):
    def available_secret_keys():
        yield ("Dlique", DLIQUE_SECRET_KEY)

    secret_sharing = secret_sharing_2_groups_required_with_1_minimum_each
    encrypted_shares = secret_sharing.generate_encrypted_shares(
        "2-groups-required-with-1-minimum-each", OBJECT_SECRET_KEY
    )
    with pytest.raises(SecretRecoveryError):
        _ = recover_object_decryption_key_from_encrypted_shares(
            encrypted_shares, available_secret_keys
        )


def test_object_decryption_key_recovery_with_known_shares(
    secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each,
):
    def available_secret_keys():
        yield ("Essun", ESSUN_SECRET_KEY)

    secret_sharing = (
        secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each
    )
    encrypted_shares = secret_sharing.generate_encrypted_shares(
        "2-groups-required-of-3-with-1-and-two-minimum-in-each", OBJECT_SECRET_KEY
    )
    camille_mnemonic = age_decrypt(
        CAMILLE_SECRET_KEY, encrypted_shares["Camille"]
    ).decode("us-ascii")
    bob_mnemonic = age_decrypt(BOB_SECRET_KEY, encrypted_shares["Bob"]).decode(
        "us-ascii"
    )
    recovered_key = recover_object_decryption_key_from_encrypted_shares(
        encrypted_shares,
        available_secret_keys,
        known_mnemonics=[
            camille_mnemonic,
            bob_mnemonic,
        ],
    )
    assert recovered_key == OBJECT_SECRET_KEY


def test_secret_sharing_errors_with_duplicate_identifiers():
    conf = f"""\
        minimum_required_groups: 2
        groups:
          legal:
            minimum_required_shares: 1
            recipient_keys:
              "Ali": {ALI_PUBLIC_KEY}
              "Bob": {BOB_PUBLIC_KEY}
          sysadmins:
            minimum_required_shares: 1
            recipient_keys:
              # Identifier also present in the legal group
              "Ali": {CAMILLE_PUBLIC_KEY}
              "Dlique": {DLIQUE_PUBLIC_KEY}
    """
    with pytest.raises(ValueError, match="Duplicate share identifier"):
        _ = SecretSharing.from_dict(yaml.safe_load(conf))


def test_secret_sharing_errors_with_duplicate_keys():
    conf = f"""\
        minimum_required_groups: 2
        groups:
          legal:
            minimum_required_shares: 1
            recipient_keys:
              "Ali": {ALI_PUBLIC_KEY}
              "Bob": {BOB_PUBLIC_KEY}
          sysadmins:
            minimum_required_shares: 1
            recipient_keys:
              # Recipient key is the same as Ali’s
              "Camille": {ALI_PUBLIC_KEY}
              "Dlique": {DLIQUE_PUBLIC_KEY}
    """
    with pytest.raises(ValueError, match="Duplicate recipient public key"):
        _ = SecretSharing.from_dict(yaml.safe_load(conf))


MOCK_AGE_PLUGIN_YUBIKEY_SCRIPT = """\
#!/bin/sh

if [ "$1" != "--identity" ]; then
    exit 1
fi

cat <<EOF
#       Serial: 5229836, Slot: 1
#         Name: age identity c3fa08e1
#      Created: Wed, 28 Jun 2023 12:24:19 +0000
#   PIN policy: Never  (A PIN is NOT required to decrypt)
# Touch policy: Cached (A physical touch is required for decryption, and is cached for 15 seconds)
#    Recipient: age1yubikey1q2e37f74zzazz75mtggzql3at66pegemfnul0dtd7axctahljkvsqezscaq
AGE-PLUGIN-YUBIKEY-1PNX57QYZC0AQ3CGZL7UL3


#       Serial: 4245067, Slot: 1
#         Name: age identity 4c1bc5f1
#      Created: Wed, 28 Jun 2023 14:43:01 +0000
#   PIN policy: Never  (A PIN is NOT required to decrypt)
# Touch policy: Always (A physical touch is required for every decryption)
#    Recipient: age1yubikey1q04k5fs8cz6kypt7vjetl2gc2qtpz9lyzpxrvv2agzt6h8n3awmzk9sgd8v
AGE-PLUGIN-YUBIKEY-1F0RYQQYZFSDUTUG4RNAMD
EOF
"""  # noqa: B950


@pytest.fixture
def mock_age_yubikey_plugin(tmp_path):
    script_path = tmp_path / "age-plugin-yubikey"
    script_path.write_text(MOCK_AGE_PLUGIN_YUBIKEY_SCRIPT)
    os.chmod(script_path, 0o500)
    old_path = os.environ["PATH"]
    os.environ["PATH"] = f"{tmp_path}:{os.environ['PATH']}"
    yield
    os.environ["PATH"] = old_path


def test_list_yubikey_identities(mock_age_yubikey_plugin):
    assert list_yubikey_identities() == [
        ("YubiKey serial 5229836 slot 1", "AGE-PLUGIN-YUBIKEY-1PNX57QYZC0AQ3CGZL7UL3"),
        ("YubiKey serial 4245067 slot 1", "AGE-PLUGIN-YUBIKEY-1F0RYQQYZFSDUTUG4RNAMD"),
    ]


@pytest.fixture
def encrypted_shares_for_object_private_key(
    secret_sharing_2_groups_required_with_1_minimum_each,
):
    secret_sharing = secret_sharing_2_groups_required_with_1_minimum_each
    return secret_sharing.generate_encrypted_shares(
        "2-groups-with-1-minimum-each", OBJECT_SECRET_KEY
    )


def test_create_recovery_bundle(
    tmp_path,
    sample_populated_storage_with_matching_hash,
    encrypted_shares_for_object_private_key,
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    swhids = [
        # Content
        "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
        "swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea",
        # SkippedContent
        "swh:1:cnt:33e45d56f88993aae6a0198013efa80716fd8920",
        # Directory
        "swh:1:dir:5256e856a0a0898966d6ba14feb4388b8b82d302",
        "swh:1:dir:4b825dc642cb6eb9a060e54bf8d69288fbee4904",
        "swh:1:dir:afa0105cfcaa14fdbacee344e96659170bb1bda5",
        # Revision
        "swh:1:rev:01a7114f36fddd5ef2511b2cadda237a68adbb12",
        "swh:1:rev:a646dd94c912829659b22a1e7e143d2fa5ebde1b",
        # Release
        "swh:1:rel:f7f222093a18ec60d781070abec4a630c850b837",
        "swh:1:rel:db81a26783a3f4a9db07b4759ffc37621f159bb2",
        # Snapshot
        "swh:1:snp:9b922e6d8d5b803c1582aabe5525b7b91150788e",
        "swh:1:snp:db99fda25b43dc5cd90625ee4b0744751799c917",
        # Origin
        "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645",
        "swh:1:ori:9147ab9c9287940d4fdbe95d8780664d7ad2dfc0",
        # RawExtrinsicMetadata
        "swh:1:emd:101d70c3574c1e4b730d7ba8e83a4bdadc8691cb",
        "swh:1:emd:ef3b0865c7a05f79772a3189ddfc8515ec3e1844",
        "swh:1:emd:43dad4d96edf2fb4f77f0dbf72113b8fe8b5b664",
        "swh:1:emd:9cafd9348f3a7729c2ef0b9b149ba421589427f0",
    ]

    unique_keys_found = []

    def register(obj):
        if hasattr(obj, "swhid"):
            unique_keys_found.append(obj.swhid().object_id)
        else:
            unique_keys_found.append(obj.unique_key())

    with RecoveryBundleCreator(
        path=bundle_path,
        storage=sample_populated_storage_with_matching_hash,
        removal_identifier="test_bundle",
        requested=[
            Origin("https://github.com/user1/repo1"),
            Origin("https://github.com/user2/repo1"),
        ],
        referencing=[
            ExtendedSWHID.from_string(s)
            for s in (
                "swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa",
                "swh:1:dir:8505808532953da7d2581741f01b29c04b1cb9ab",
            )
        ],
        object_public_key=OBJECT_PUBLIC_KEY,
        decryption_key_shares=encrypted_shares_for_object_private_key,
        registration_callback=register,
    ) as creator:
        creator.backup_swhids(ExtendedSWHID.from_string(swhid) for swhid in swhids)

    from zipfile import ZipFile

    from swh.journal.serializers import kafka_to_value

    with ZipFile(bundle_path, "r") as bundle:
        # Do we have the expected files?
        assert bundle.namelist() == [
            "skipped_contents/swh_1_cnt_33e45d56f88993aae6a0198013efa80716fd8920_1.age",
            "contents/swh_1_cnt_c932c7649c6dfa4b82327d121215116909eb3bea.age",
            "contents/swh_1_cnt_d81cc0710eb6cf9efd5b920a8453e1e07157b6cd.age",
            "directories/swh_1_dir_4b825dc642cb6eb9a060e54bf8d69288fbee4904.age",
            "directories/swh_1_dir_5256e856a0a0898966d6ba14feb4388b8b82d302.age",
            "directories/swh_1_dir_afa0105cfcaa14fdbacee344e96659170bb1bda5.age",
            "extids/fa730cf0bb415e1e921e430984bdcddd9c8eea4a.age",
            "revisions/swh_1_rev_01a7114f36fddd5ef2511b2cadda237a68adbb12.age",
            "extids/486e20ccedc221075b12abbb607a888875db41f6.age",
            "revisions/swh_1_rev_a646dd94c912829659b22a1e7e143d2fa5ebde1b.age",
            "releases/swh_1_rel_db81a26783a3f4a9db07b4759ffc37621f159bb2.age",
            "releases/swh_1_rel_f7f222093a18ec60d781070abec4a630c850b837.age",
            "snapshots/swh_1_snp_9b922e6d8d5b803c1582aabe5525b7b91150788e.age",
            "snapshots/swh_1_snp_db99fda25b43dc5cd90625ee4b0744751799c917.age",
            "origins/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645.age",
            "origin_visits/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645_1.age",
            "origin_visit_statuses/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645_1_2015-01-01T23_00_00.000000+00_00.age",
            "origin_visits/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645_2.age",
            "origin_visit_statuses/swh_1_ori_33abd4b4c5db79c7387673f71302750fd73e0645_2_2017-01-01T23_00_00.000000+00_00.age",
            "origins/swh_1_ori_9147ab9c9287940d4fdbe95d8780664d7ad2dfc0.age",
            "origin_visits/swh_1_ori_9147ab9c9287940d4fdbe95d8780664d7ad2dfc0_1.age",
            "origin_visit_statuses/swh_1_ori_9147ab9c9287940d4fdbe95d8780664d7ad2dfc0_1_2015-01-01T23_00_00.000000+00_00.age",
            "raw_extrinsic_metadata/1_swh_1_emd_101d70c3574c1e4b730d7ba8e83a4bdadc8691cb.age",
            "raw_extrinsic_metadata/2_swh_1_emd_ef3b0865c7a05f79772a3189ddfc8515ec3e1844.age",
            "raw_extrinsic_metadata/3_swh_1_emd_43dad4d96edf2fb4f77f0dbf72113b8fe8b5b664.age",
            "raw_extrinsic_metadata/4_swh_1_emd_9cafd9348f3a7729c2ef0b9b149ba421589427f0.age",
            "manifest.yml",
        ]
        # Can we load the manifest?
        manifest = Manifest.load(bundle.read("manifest.yml"))
        assert isinstance(manifest, Manifest)
        assert manifest.removal_identifier == "test_bundle"
        assert (
            ExtendedSWHID.from_string(
                "swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa"
            )
            in manifest.referencing
        )
        # Can we unpack, decrypt and load at least an object?
        encrypted_serialized_content = bundle.read(
            "contents/swh_1_cnt_d81cc0710eb6cf9efd5b920a8453e1e07157b6cd.age"
        )
        serialized_content = age_decrypt(
            OBJECT_SECRET_KEY, encrypted_serialized_content
        )
        content: Content = Content.from_dict(kafka_to_value(serialized_content))
        assert content.status == "visible"
        # Have we properly saved content data?
        assert content.data == b"42\n"
        # Have we registered all saved objects?
        assert unique_keys_found == [
            bytes.fromhex("33e45d56f88993aae6a0198013efa80716fd8920"),
            bytes.fromhex("c932c7649c6dfa4b82327d121215116909eb3bea"),
            bytes.fromhex("d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"),
            bytes.fromhex("4b825dc642cb6eb9a060e54bf8d69288fbee4904"),
            bytes.fromhex("5256e856a0a0898966d6ba14feb4388b8b82d302"),
            bytes.fromhex("afa0105cfcaa14fdbacee344e96659170bb1bda5"),
            bytes.fromhex("fa730cf0bb415e1e921e430984bdcddd9c8eea4a"),
            bytes.fromhex("01a7114f36fddd5ef2511b2cadda237a68adbb12"),
            bytes.fromhex("486e20ccedc221075b12abbb607a888875db41f6"),
            bytes.fromhex("a646dd94c912829659b22a1e7e143d2fa5ebde1b"),
            bytes.fromhex("db81a26783a3f4a9db07b4759ffc37621f159bb2"),
            bytes.fromhex("f7f222093a18ec60d781070abec4a630c850b837"),
            bytes.fromhex("9b922e6d8d5b803c1582aabe5525b7b91150788e"),
            bytes.fromhex("db99fda25b43dc5cd90625ee4b0744751799c917"),
            bytes.fromhex("33abd4b4c5db79c7387673f71302750fd73e0645"),
            {
                "date": "2015-01-01 23:00:00+00:00",
                "origin": "https://github.com/user1/repo1",
            },
            {
                "date": "2015-01-01 23:00:00+00:00",
                "origin": "https://github.com/user1/repo1",
                "visit": "1",
            },
            {
                "date": "2017-01-01 23:00:00+00:00",
                "origin": "https://github.com/user1/repo1",
            },
            {
                "date": "2017-01-01 23:00:00+00:00",
                "origin": "https://github.com/user1/repo1",
                "visit": "2",
            },
            bytes.fromhex("9147ab9c9287940d4fdbe95d8780664d7ad2dfc0"),
            {
                "date": "2015-01-01 23:00:00+00:00",
                "origin": "https://github.com/user2/repo1",
            },
            {
                "date": "2015-01-01 23:00:00+00:00",
                "origin": "https://github.com/user2/repo1",
                "visit": "1",
            },
            bytes.fromhex("101d70c3574c1e4b730d7ba8e83a4bdadc8691cb"),
            bytes.fromhex("ef3b0865c7a05f79772a3189ddfc8515ec3e1844"),
            bytes.fromhex("43dad4d96edf2fb4f77f0dbf72113b8fe8b5b664"),
            bytes.fromhex("9cafd9348f3a7729c2ef0b9b149ba421589427f0"),
        ]


def test_create_recovery_bundle_fails_if_empty(
    tmp_path,
    sample_populated_storage_with_matching_hash,
    encrypted_shares_for_object_private_key,
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    with pytest.raises(ValueError):
        with RecoveryBundleCreator(
            path=bundle_path,
            storage=sample_populated_storage_with_matching_hash,
            removal_identifier="test_bundle",
            requested=[Origin("https://github.com/user1/repo1")],
            referencing=[],
            object_public_key=OBJECT_PUBLIC_KEY,
            decryption_key_shares=encrypted_shares_for_object_private_key,
        ) as _:
            # oops, we do not add any objects
            pass


def test_create_recovery_bundle_fails_without_decryption_key_shares(
    tmp_path, sample_populated_storage_with_matching_hash
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    swhids = ["swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645"]
    with pytest.raises(ValueError):
        with RecoveryBundleCreator(
            path=bundle_path,
            storage=sample_populated_storage_with_matching_hash,
            removal_identifier="test_bundle",
            requested=[Origin("https://github.com/user1/repo1")],
            referencing=[],
            object_public_key=OBJECT_PUBLIC_KEY,
            decryption_key_shares={},
        ) as creator:
            creator.backup_swhids(ExtendedSWHID.from_string(swhid) for swhid in swhids)


def test_create_recovery_bundle_with_optional_fields(
    tmp_path,
    sample_populated_storage_with_matching_hash,
    encrypted_shares_for_object_private_key,
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    swhids = ["swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645"]
    expiration_date = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(
        days=365
    )
    with RecoveryBundleCreator(
        path=bundle_path,
        storage=sample_populated_storage_with_matching_hash,
        removal_identifier="test_bundle",
        requested=[Origin("https://github.com/user1/repo1")],
        referencing=[],
        object_public_key=OBJECT_PUBLIC_KEY,
        decryption_key_shares=encrypted_shares_for_object_private_key,
    ) as creator:
        creator.backup_swhids(ExtendedSWHID.from_string(swhid) for swhid in swhids)
        creator.set_reason("we are running a test")
        creator.set_expire(expiration_date)

    from zipfile import ZipFile

    with ZipFile(bundle_path, "r") as bundle:
        manifest = Manifest.load(bundle.read("manifest.yml"))
        assert manifest.reason == "we are running a test"
        assert manifest.expire.isoformat(
            timespec="seconds"
        ) == expiration_date.isoformat(timespec="seconds")


def test_create_recovery_bundle_refuses_multiple_backup_call_with_emd(
    tmp_path,
    sample_populated_storage_with_matching_hash,
    encrypted_shares_for_object_private_key,
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    swhids = [
        ExtendedSWHID.from_string(s)
        for s in (
            # RawExtrinsicMetadata
            "swh:1:emd:101d70c3574c1e4b730d7ba8e83a4bdadc8691cb",
            "swh:1:emd:ef3b0865c7a05f79772a3189ddfc8515ec3e1844",
        )
    ]
    with RecoveryBundleCreator(
        path=bundle_path,
        storage=sample_populated_storage_with_matching_hash,
        removal_identifier="test_bundle",
        requested=[
            Origin("https://github.com/user1/repo1"),
        ],
        referencing=[],
        object_public_key=OBJECT_PUBLIC_KEY,
        decryption_key_shares=encrypted_shares_for_object_private_key,
    ) as creator:
        # We add a first batch of 1 RawExtrinsincMetadata object
        creator.backup_swhids([swhids[0]])
        # Trying to add a second batch must fail
        with pytest.raises(ValueError, match="RawExtrinsingMetdata objects"):
            creator.backup_swhids([swhids[1]])


def test_create_recovery_bundle_when_content_data_not_found(
    mocker,
    tmp_path,
    sample_populated_storage_with_matching_hash,
    encrypted_shares_for_object_private_key,
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    swhids = [
        ExtendedSWHID.from_string(s)
        for s in (
            # Content
            "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
            "swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea",
        )
    ]
    # None means data for the given hashes have not been found
    mocker.patch.object(
        sample_populated_storage_with_matching_hash,
        "content_get_data",
        return_value=None,
    )
    with pytest.raises(ContentDataNotFound):
        with RecoveryBundleCreator(
            path=bundle_path,
            storage=sample_populated_storage_with_matching_hash,
            removal_identifier="test_bundle",
            requested=[
                Origin("https://github.com/user1/repo1"),
            ],
            referencing=[],
            object_public_key=OBJECT_PUBLIC_KEY,
            decryption_key_shares=encrypted_shares_for_object_private_key,
        ) as creator:
            creator.backup_swhids(swhids)


def test_create_recovery_bundle_when_content_data_not_found_warns(
    mocker,
    caplog,
    tmp_path,
    sample_populated_storage_with_matching_hash,
    encrypted_shares_for_object_private_key,
):
    bundle_path = tmp_path / "test.swh-recovery-bundle"
    swhids = [
        ExtendedSWHID.from_string(s)
        for s in (
            # Content
            "swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea",
            "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd",
        )
    ]
    # None means data for the given hashes have not been found
    mocker.patch.object(
        sample_populated_storage_with_matching_hash,
        "content_get_data",
        return_value=None,
    )
    with caplog.at_level(logging.WARNING):
        with RecoveryBundleCreator(
            path=bundle_path,
            storage=sample_populated_storage_with_matching_hash,
            removal_identifier="test_bundle",
            requested=[
                Origin("https://github.com/user1/repo1"),
            ],
            referencing=[],
            object_public_key=OBJECT_PUBLIC_KEY,
            decryption_key_shares=encrypted_shares_for_object_private_key,
            allow_empty_content_objects=True,
        ) as creator:
            creator.backup_swhids(swhids)
    assert caplog.messages == [
        "No data available for swh:1:cnt:c932c7649c6dfa4b82327d121215116909eb3bea. "
        "Recording empty Content object as requested.",
        "No data available for swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd. "
        "Recording empty Content object as requested.",
    ]


def test_recovery_bundle_decryption_key_provider_is_optional(
    sample_recovery_bundle_path,
):
    bundle = RecoveryBundle(sample_recovery_bundle_path)
    # We can access the manifest even without a decryption_key_provider
    assert bundle.removal_identifier == "test_bundle"
    # But trying to get the object_decryption_key will fail
    with pytest.raises(ValueError):
        _ = bundle.object_decryption_key


def test_recovery_bundle_requested(request, sample_recovery_bundle):
    if "version-1" in request.keywords or "version-2" in request.keywords:
        with pytest.raises(UnsupportedFeatureException):
            _ = sample_recovery_bundle.requested
    else:
        assert sample_recovery_bundle.requested == [
            Origin(url="https://github.com/user1/repo1"),
            Origin(url="https://github.com/user2/repo1"),
        ]


def test_recovery_bundle_referencing(request, sample_recovery_bundle):
    if "version-1" in request.keywords or "version-2" in request.keywords:
        with pytest.raises(UnsupportedFeatureException):
            _ = sample_recovery_bundle.referencing
    else:
        assert sample_recovery_bundle.referencing == [
            ExtendedSWHID.from_string(
                "swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa"
            ),
            ExtendedSWHID.from_string(
                "swh:1:dir:8505808532953da7d2581741f01b29c04b1cb9ab"
            ),
        ]


def test_recovery_bundle_get_missing_referenced_objects(
    request, sample_recovery_bundle, swh_storage, sample_data
):
    if "version-1" in request.keywords or "version-2" in request.keywords:
        with pytest.raises(UnsupportedFeatureException):
            _ = sample_recovery_bundle.referencing
        return

    # swh_storage is empty at this point
    assert sample_recovery_bundle.get_missing_referenced_objects(swh_storage) == {
        ExtendedSWHID.from_string(s)
        for s in (
            "swh:1:cnt:36fade77193cb6d2bd826161a0979d64c28ab4fa",
            "swh:1:dir:8505808532953da7d2581741f01b29c04b1cb9ab",
        )
    }
    # now load these objects
    swh_storage.content_add([sample_data.content2])
    swh_storage.directory_add([sample_data.directory2])
    result = swh_storage.flush()
    assert result["content:add"] == 1
    assert result["directory:add"] == 1
    # none should be missing
    assert sample_recovery_bundle.get_missing_referenced_objects(swh_storage) == set()


def test_recovery_bundle_get_dict(sample_recovery_bundle):
    swhid = ExtendedSWHID.from_string(
        "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645"
    )
    d = sample_recovery_bundle.get_dict(swhid)
    assert d == {"url": "https://github.com/user1/repo1"}


def test_recovery_bundle_write_content_data(tmp_path, sample_recovery_bundle):
    dest_path = tmp_path / "data"
    swhid = ExtendedSWHID.from_string(
        "swh:1:cnt:d81cc0710eb6cf9efd5b920a8453e1e07157b6cd"
    )
    with open(dest_path, "wb") as dest:
        sample_recovery_bundle.write_content_data(swhid, dest)
    assert open(dest_path, "rb").read() == b"42\n"


def sorted_by_swhid(objs):
    def key(obj) -> str:
        return str(obj.swhid())

    return sorted(objs, key=key)


def test_recovery_bundle_contents(sample_recovery_bundle, sample_data):
    contents = list(sample_recovery_bundle.contents())
    assert contents == sorted_by_swhid([sample_data.content, sample_data.content3])


def test_recovery_bundle_skipped_contents(sample_recovery_bundle, sample_data):
    skipped_contents = list(sample_recovery_bundle.skipped_contents())
    assert len(skipped_contents) == 1
    skipped_content = skipped_contents[0]
    # None-ify fields that are invalid in reference data
    skipped_content = attr.evolve(skipped_content, ctime=None, origin=None)
    ref_skipped_content = attr.evolve(
        sample_data.skipped_content, ctime=None, origin=None
    )
    assert skipped_content == ref_skipped_content


def test_recovery_bundle_directories(sample_recovery_bundle, sample_data):
    directories = list(sample_recovery_bundle.directories())
    # Comparing Directory objects directly will not work because order
    # of entries are not guaranteed by storage. Therefore we have to
    # compare each field individually.
    assert len(directories) == 3
    for d, sample_d in zip(
        directories,
        sorted_by_swhid(
            [sample_data.directory, sample_data.directory5, sample_data.directory6]
        ),
    ):
        assert d.id == sample_d.id
        assert set(d.entries) == set(sample_d.entries)
        assert d.raw_manifest == sample_d.raw_manifest


def test_recovery_bundle_revisions(sample_recovery_bundle, sample_data):
    revisions = list(sample_recovery_bundle.revisions())
    assert revisions == sorted_by_swhid([sample_data.revision, sample_data.revision2])


def test_recovery_bundle_releases(sample_recovery_bundle, sample_data):
    releases = list(sample_recovery_bundle.releases())
    assert releases == sorted_by_swhid([sample_data.release, sample_data.release2])


def test_recovery_bundle_snapshots(sample_recovery_bundle, sample_data):
    snapshots = list(sample_recovery_bundle.snapshots())
    assert snapshots == sorted_by_swhid(
        [sample_data.snapshot, sample_data.complete_snapshot]
    )


def test_recovery_bundle_origins(sample_recovery_bundle, sample_data):
    origins = list(sample_recovery_bundle.origins())
    assert origins == sorted_by_swhid([sample_data.origin, sample_data.origin2])


def test_recovery_bundle_origin_visits(sample_recovery_bundle, sample_data):
    origin_visits = list(sample_recovery_bundle.origin_visits(sample_data.origin))
    assert origin_visits == [
        sample_data.origin_visit,
        sample_data.origin_visit2,
    ]


def test_recovery_bundle_origin_visit_statuses(sample_recovery_bundle, sample_data):
    origin_visit_statuses = list(
        sample_recovery_bundle.origin_visit_statuses(sample_data.origin)
    )
    assert origin_visit_statuses == [
        sample_data.origin_visit_status,
        sample_data.origin_visit2_status,
    ]


def test_recovery_bundle_raw_extrinsic_metdata(sample_recovery_bundle, sample_data):
    if sample_recovery_bundle.version < 2:
        pytest.skip("old bundle does not contain RawExtrinsicMetadata objects")
    emds = list(sample_recovery_bundle.raw_extrinsic_metadata())
    assert emds == [
        sample_data.content_metadata1,
        sample_data.content_metadata2,
        sample_data.origin_metadata1,
        sample_data.origin_metadata2,
    ]


def test_recovery_bundle_extids(sample_recovery_bundle, sample_data):
    if sample_recovery_bundle.version < 2:
        pytest.skip("old bundle does not contain ExtID objects")
    extids = list(sample_recovery_bundle.extids())
    assert extids == [sample_data.extid1, sample_data.extid3]


def test_restore(sample_recovery_bundle, swh_storage, sample_data):
    expected_result = {
        "content:add": 2,
        "content:add:bytes": 10,
        "skipped_content:add": 1,
        "directory:add": 3,
        "revision:add": 2,
        "release:add": 2,
        "snapshot:add": 2,
        "origin:add": 2,
        "origin_visit:add": 3,
        "origin_visit_status:add": 3,
    }
    if sample_recovery_bundle.version >= 2:
        expected_result |= {
            "ori_metadata:add": 2,
            "cnt_metadata:add": 2,
            "extid:add": 2,
        }
        # We are going to restore RawExtrinsicMetadata objects
        # which requires the right MetadataAuthority and MetadataFetcher
        # objects to be present in storage. So we need to load them first.
        # This is ok for now as we don’t remove any MetadataAuthority
        # or MetadataFetcher, but this might change in the future.
        # Issue tracked as:
        # https://gitlab.softwareheritage.org/swh/devel/swh-alter/-/issues/21
        swh_storage.metadata_authority_add(sample_data.authorities)
        swh_storage.metadata_fetcher_add(sample_data.fetchers)
    result = sample_recovery_bundle.restore(swh_storage)
    # Remove the “object_reference:add” entry if it exists as we don’t care
    _ = result.pop("object_reference:add", None)
    assert result == expected_result
    [
        origin,
    ] = swh_storage.origin_get(["https://github.com/user1/repo1"])
    assert origin is not None
    assert str(origin.swhid()) == "swh:1:ori:33abd4b4c5db79c7387673f71302750fd73e0645"
    origin_visit = swh_storage.origin_visit_get_by(origin.url, 1)
    assert origin_visit == sample_data.origin_visit
    origin_visit2 = swh_storage.origin_visit_get_by(origin.url, 2)
    assert origin_visit2 == sample_data.origin_visit2
    origin_visit2_status = swh_storage.origin_visit_status_get_latest(origin.url, 2)
    assert origin_visit2_status == sample_data.origin_visit2_status


def test_rollover(
    tmp_path,
    sample_recovery_bundle_path,
    secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each,
):
    import shutil

    bundle_path = shutil.copy(
        sample_recovery_bundle_path, tmp_path / "rollover.swh-recovery-bundle"
    )
    secret_sharing = (
        secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each
    )
    bundle = RecoveryBundle(bundle_path, object_decryption_key_provider_for_sample)

    # Record OriginVisit and OriginVisitStatuses objects for later comparison
    origin_visits = set()
    origin_visit_statuses = set()
    for origin in bundle.origins():
        origin_visits.update(bundle.origin_visits(origin))
        origin_visit_statuses.update(bundle.origin_visit_statuses(origin))

    # Perform rollover!
    bundle.rollover(secret_sharing)

    # Check all the things!!
    def decryption_key_provider(manifest):
        return recover_object_decryption_key_from_encrypted_shares(
            manifest.decryption_key_shares,
            available_secret_keys_for_2_groups_required_of_3_one_and_two,
        )

    new_bundle = RecoveryBundle(bundle_path, decryption_key_provider)
    # Do we have the new holders?
    assert new_bundle.share_ids == {"Ali", "Bob", "Camille", "Dlique", "Essun"}
    # Has the old object been updated as well?
    assert bundle.share_ids == new_bundle.share_ids
    # Is the decryption key still the same after being recovered?
    assert new_bundle.object_decryption_key == OBJECT_SECRET_KEY
    # Can we still decrypt all known objects?
    decrypted_swhids = set()
    for obj in itertools.chain(
        new_bundle.contents(),
        new_bundle.skipped_contents(),
        new_bundle.directories(),
        new_bundle.revisions(),
        new_bundle.releases(),
        new_bundle.snapshots(),
    ):
        decrypted_swhids.add(obj.swhid().to_extended())
    new_origin_visits = set()
    new_origin_visit_statuses = set()
    for origin in new_bundle.origins():
        decrypted_swhids.add(origin.swhid())
        new_origin_visits.update(new_bundle.origin_visits(origin))
        new_origin_visit_statuses.update(new_bundle.origin_visit_statuses(origin))
    if bundle.version >= 2:
        decrypted_swhids.update(
            [emd.swhid() for emd in new_bundle.raw_extrinsic_metadata()]
        )
        new_bundle.extids() == bundle.extids()
    assert decrypted_swhids == set(bundle.swhids)
    assert origin_visits == new_origin_visits
    assert origin_visit_statuses == new_origin_visit_statuses


def test_rollover_fails_when_unable_to_write(
    tmp_path,
    sample_recovery_bundle_path,
    secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each,
):
    import shutil

    bundle_path = shutil.copy(
        sample_recovery_bundle_path, tmp_path / "rollover.swh-recovery-bundle"
    )
    secret_sharing = (
        secret_sharing_2_groups_required_of_3_with_1_and_two_minimum_in_each
    )
    bundle = RecoveryBundle(bundle_path, object_decryption_key_provider_for_sample)
    share_ids = bundle.share_ids
    swhids = bundle.swhids

    # We can more or less simulate a full disk by preventing writes
    # in the directory holding the recovery bundle.
    os.chmod(tmp_path, 0o500)

    with pytest.raises(OSError):
        bundle.rollover(secret_sharing)

    new_bundle = RecoveryBundle(bundle_path, object_decryption_key_provider_for_sample)
    assert new_bundle.share_ids == share_ids
    assert new_bundle.object_decryption_key == OBJECT_SECRET_KEY
    assert new_bundle.get_dict(swhids[0]) is not None



# Copyright (C) 2024 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import logging
from types import TracebackType
from typing import (
    Callable,
    Generic,
    Iterable,
    Iterator,
    Optional,
    Protocol,
    TypeVar,
    cast,
)

V = TypeVar("V")

logger = logging.getLogger(__name__)


class ProgressBar(Protocol, Generic[V]):
    """Interface for the ProgressBar object, mimicking click’s."""

    def __enter__(self) -> "ProgressBar[V]": ...

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None: ...

    def __iter__(self) -> Iterator[V]: ...

    def __next__(self) -> V: ...

    def update(self, n_steps: int, current_item: Optional[V] = None) -> None: ...


class NoProgressBar(Generic[V]):
    """A ProgressBar implementation that displays nothing.

    Typically returned by :py:func:`no_progressbar`.
    """

    def __init__(self, iterable: Iterable[V], label: Optional[str] = None):
        self.iter = iter(iterable)
        if label:
            logger.info(label)

    def __enter__(self) -> "NoProgressBar[V]":
        return self

    def __exit__(
        self,
        exc_type: Optional[type[BaseException]],
        exc_value: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        pass

    def __iter__(self) -> Iterator[V]:
        yield from self.iter

    def __next__(self) -> V:
        return next(iter(self))

    def update(self, n_steps: int, current_item: Optional[V] = None) -> None:
        pass


class ProgressBarInit(Protocol):
    """A protocol abstracting the ``click.progressbar()`` function."""

    def __call__(
        self,
        iterable: Optional[Iterable[V]] = None,
        length: Optional[int] = None,
        label: Optional[str] = None,
        show_eta: bool = True,
        show_pos: bool = False,
        show_percent: Optional[bool] = None,
        item_show_func: Optional[Callable[[V], str]] = None,
    ) -> ProgressBar[V]: ...


def no_progressbar(
    iterable: Optional[Iterable[V]] = None,
    length: Optional[int] = None,
    label: Optional[str] = None,
    show_eta: bool = True,
    show_pos: bool = False,
    show_percent: Optional[bool] = None,
    item_show_func: Optional[Callable[[V], str]] = None,
) -> ProgressBar[V]:
    """Returns a :py:class:`ProgressBar` that displays nothing.

    This function allows to use the same code structure wherever a progressbar
    has to be displayed or not."""

    if iterable is not None:
        return NoProgressBar(iterable, label=label)
    elif length is not None:
        # Without this `cast()`, mypy thinks we return a `ProgressBar[int]`.
        # While true, it only happens in the case that V has not been specified,
        # so we are in our rights to state that V=int this time being.
        return NoProgressBar(cast(Iterable[V], iter(range(0, length))), label=label)
    else:
        raise ValueError("Either `iterable or `length` must be specified.")



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
This module implements the marking stage of the
:ref:`removal algorithm <alter_removal_algorithm>`.
"""

from enum import Enum, auto
from functools import partial
from itertools import chain
import logging
from typing import Iterator, List, Optional, Set

from igraph import Vertex

from swh.graph.http_client import GraphArgumentException, RemoteGraphClient
from swh.model.swhids import ExtendedObjectType as ObjectType
from swh.model.swhids import ExtendedSWHID
from swh.storage.interface import StorageInterface

from .inventory import InventorySubgraph
from .progressbar import ProgressBar, ProgressBarInit, no_progressbar
from .subgraph import Subgraph
from .utils import get_filtered_objects

logger = logging.getLogger(__name__)

NEVER_REMOVABLE: Set[ExtendedSWHID] = {
    ExtendedSWHID.from_string(swhid)
    for swhid in [
        # Empty content
        "swh:1:cnt:e69de29bb2d1d6434b8b29ae775ad8c2e48c5391",
        # 1-byte contents
        "swh:1:cnt:0015308348bf55e30efa51f842ae48361a8d0bdb",
        "swh:1:cnt:0022a3ee843f9c3d983aaf4f255cefeafe2859a0",
        "swh:1:cnt:009080e8e0bf834027bd7a55094c2e09e868e67b",
        "swh:1:cnt:00b15c0a321af2492988194cb33b5f55b2ae8332",
        "swh:1:cnt:013d565bb4055b02b0853816a9f6c69d1478f3f3",
        "swh:1:cnt:02358d2358658574ba0767140caa4216ee7ea5bf",
        "swh:1:cnt:02691e3522cd51ad600902451eace5b54cce5018",
        "swh:1:cnt:04f7b5be6987d303c24a0aa2761207922f6bbf66",
        "swh:1:cnt:050ac90ecbd9ce5a88212058fad711b4231d104d",
        "swh:1:cnt:0519ecba6ea913e21689ec692e81e9e4973fbf73",
        "swh:1:cnt:05de4782e137f9a25e63fa2cac3c94ba218dc69d",
        "swh:1:cnt:05fdf94240ddb18dae9aa42f4d35d6099b93bc7e",
        "swh:1:cnt:07e2dac48ff656acc86895efc33941223195b371",
        "swh:1:cnt:080c324fec6abbac88b03ca1298b71b6d69f9de1",
        "swh:1:cnt:0817502b11d3d776cac53346c03336e236cfa665",
        "swh:1:cnt:083b700b23248d19fdf63d7501b802a9556dcd90",
        "swh:1:cnt:08b9811c98f0d90dbacc006ddcd80c5945b9ea55",
        "swh:1:cnt:08e5b85ed7d87899de10f8cda7f7d94c8541a8fd",
        "swh:1:cnt:0901ffdd0da48f1648b19e1833a4e2d9b5a24fbe",
        "swh:1:cnt:0d758c9c7bc06c1e307f05d92d896aaf0a8a6d2c",
        "swh:1:cnt:0f0f3e3b1ff2bc6722afc3e3812e6b782683896f",
        "swh:1:cnt:0f137124110a2cc080e70794f98f7c8e5fd87e75",
        "swh:1:cnt:0fe2fa50e8e79f56b23267be86bde68d6f86195b",
        "swh:1:cnt:12183abc9e3e34700451537045dd9c0b6db609f3",
        "swh:1:cnt:127620c78a33b65b9b1b3f40843e7236e50e8c6c",
        "swh:1:cnt:13c603768561c8efce4659491508b21865b635f8",
        "swh:1:cnt:147efaa6d93d25dca26529dc3940f2079c330880",
        "swh:1:cnt:148a7b00c1946b6655f525fcfbd1dade6cf0dd2e",
        "swh:1:cnt:14ceb3ee12a1980c71ee530968609e528159164e",
        "swh:1:cnt:152f9ed5aa2d840c4a115d34ac0271262ca416ef",
        "swh:1:cnt:1578e7b311acfb74c9dffb8a38edfa9a4171fdcc",
        "swh:1:cnt:16e0e90df089debac429ebcae3cd8ea4b886aa62",
        "swh:1:cnt:16e45d390712388410556b522c74a11637716844",
        "swh:1:cnt:195ee38114e6c3445dd9b88f3dd21cc184662067",
        "swh:1:cnt:1987bbd3f002a19fb61286144bb143274285f89d",
        "swh:1:cnt:1bbf238cfb853e6bc5f6f5c85118d6b850b581ff",
        "swh:1:cnt:1bd8bf6057312efa3c5e9d7e11fdacd688640b6c",
        "swh:1:cnt:1c8a0e7976207fb9f03ed7e260950b62b8b9d396",
        "swh:1:cnt:1d2f01491f783c8c7f0917cc68526c6307d80e39",
        "swh:1:cnt:1d79949788956cdcd635f20bc67797e63beb8df9",
        "swh:1:cnt:1d8c7673ef786c5ff5e3b3defd0040af843bbc95",
        "swh:1:cnt:1db515f9a1604d126df3ed98ffa4cc5a844b96a2",
        "swh:1:cnt:1de00ecd4fc96b12f0281d2a802ff1ecdf76256a",
        "swh:1:cnt:2105af1f86eb44a91698fd33caccf5e105b36f9e",
        "swh:1:cnt:22aac29bb31be88e6db5e2f285243c3ab0c921dc",
        "swh:1:cnt:22ded55aa2c14af167996451c06c190f270e78a9",
        "swh:1:cnt:23fa7d31a729cb3b60694cf15e906aee5823b96f",
        "swh:1:cnt:24de910c13bb1e60fc5ec37a1058d356b1f2fa4d",
        "swh:1:cnt:25cb955ba23571f6fefd57cecf5f67454210bbc8",
        "swh:1:cnt:265ee26ceb272945f7c08a8f37c79bc6c55bb4d8",
        "swh:1:cnt:2725bca0006db42c8ee38c15dfa290bbe57f9a94",
        "swh:1:cnt:27f1d22a655c42c5a5fc9ffd583789950da11fc8",
        "swh:1:cnt:2882b1818ef42e381f47427af8f2cf8dc2337e10",
        "swh:1:cnt:2a2a526d8c3f95e6b1c615a3e056806cbf8cc7a3",
        "swh:1:cnt:2b8e128b91d3f449a15bf506aa87126c40bc68d8",
        "swh:1:cnt:2e65efe2a145dda7ee51d1741299f848e5bf752e",
        "swh:1:cnt:2f259b79aa7e263f5829bb6e98096e7ec976d998",
        "swh:1:cnt:2f94675b7cc5772325e379fc48538156958bab62",
        "swh:1:cnt:301160a93062df23030a69f4b5e4d9bf71866ee9",
        "swh:1:cnt:301792a3a0f42258c745f18acbbcc5dcf0286be0",
        "swh:1:cnt:303e398c82e88040b372475614e4fe2888bcbe56",
        "swh:1:cnt:31f442a2f86c480aa94912d5c2ead0b70426e2fb",
        "swh:1:cnt:32f64f4d836716819dc5fa9a1e09a29b428881df",
        "swh:1:cnt:3410062ba67c5ed59b854387a8bc0ec012479368",
        "swh:1:cnt:34164076ecce82a01452f1c2979bac48c45e3afb",
        "swh:1:cnt:35ec3b9d7586b46c0fd3450ba21e30ef666cfcd6",
        "swh:1:cnt:3818ded670e4fc852dd922ff3299908102068e42",
        "swh:1:cnt:384f448f22e6a76511824c5395336eb84e13959f",
        "swh:1:cnt:395831125cab5cd356b62c6d4883454a8a511a28",
        "swh:1:cnt:39e8d66025179ed66f49e6d776fe9c4468d14eff",
        "swh:1:cnt:3a6e607aa5ab4cb455d463049b8c5d7ed0d6e4b8",
        "swh:1:cnt:3cf20d57b0b8258463711cedd592007b0b5cdfe8",
        "swh:1:cnt:3ea63c2ccdc40ee588966429fe688570ccb02c7c",
        "swh:1:cnt:3f1695f161395210e9ab0e22d48014e179514e21",
        "swh:1:cnt:40a2dbb37b14d86d8da15ba799af13f013722c2a",
        "swh:1:cnt:41622b472098a6142f0225b50140189cfbd51779",
        "swh:1:cnt:4238428a9eacce28e3248da18e7847c756668e6b",
        "swh:1:cnt:4287ca8617970fa8fc025b75cb319c7032706910",
        "swh:1:cnt:4489a65003548ac3a08ee5d03b9535c7f7f7dead",
        "swh:1:cnt:449e49efc2a6de36206ab0e4f38ecfce9ec15632",
        "swh:1:cnt:45a8ca02bfc82ca6f053c14f35bc90a6382b2612",
        "swh:1:cnt:45c30f49c0daf7d28dedf8e6caf6196e5caf4a92",
        "swh:1:cnt:46df076df476712741056c452b634482efcf74d8",
        "swh:1:cnt:4977bc62c0e5eedda3a63e1250139237f32fb671",
        "swh:1:cnt:4be2460889810589fe96270d93248b528079cc7b",
        "swh:1:cnt:4d1ae35ba2c8ec712fa2a379db44ad639ca277bd",
        "swh:1:cnt:4e3bc519eef7d5a35fff67687eaee65616832e45",
        "swh:1:cnt:4f0734cbe3ae9d87287203f06e908edec7c8ed4d",
        "swh:1:cnt:4f6c4ee9d928270b4304e3abcd8d81df3e740d12",
        "swh:1:cnt:4fc3fe1ce587c1fb2acc24b2074dac4438d1b30b",
        "swh:1:cnt:500c0709ca24338426091ca19777e13a1920ebdf",
        "swh:1:cnt:501a6bbaf1e2cec78ba2c39f1bd551a43638094b",
        "swh:1:cnt:50c8be35f7782588f82895b7f1890f5fca711866",
        "swh:1:cnt:52771883833a931bf221105e2eb19fdc30a1631a",
        "swh:1:cnt:52d0524746f70b293da3def801a5823b8b9ce3a1",
        "swh:1:cnt:52e60b4495af680809173f50b2918757eb176e76",
        "swh:1:cnt:5416677bc7dab0c8bec3f5bf44d7d28b4ff73b13",
        "swh:1:cnt:54a81dcac6cfc078fbf4b74de360a4a2d9c762f5",
        "swh:1:cnt:54caf60b13678be3b856d9d9d5b1aa16b6cbe8fe",
        "swh:1:cnt:5639b6ddcf62a692b4cbd6d97d3138829c83b9e7",
        "swh:1:cnt:566d900abb9af2df549e573a5c0b8c5d7c3204fc",
        "swh:1:cnt:56a6051ca2b02b04ef92d5150c9ef600403cb1de",
        "swh:1:cnt:57814dd0025fca1ddd1e1058fa08ca9e56651f04",
        "swh:1:cnt:597a6db294cb721d184d6f12560dfa9e8a67de33",
        "swh:1:cnt:59cd9bddc043471be2a00f84e2a122af85fe8a08",
        "swh:1:cnt:5a77f05831a38ab834f5f49342e29f931e073f5a",
        "swh:1:cnt:5bd7dea14b4c665b222ff69cbb7a0496dc3b71b1",
        "swh:1:cnt:5cd813e5c5f312673ce9cf39fb832fb2d55116cc",
        "swh:1:cnt:5d33882543d5ade31a988fd3b8aef02faf12d742",
        "swh:1:cnt:60a89ed235449c0bc49f2280ca81ccc88125fff6",
        "swh:1:cnt:62f9457511f879886bb7728c986fe10b0ece6bcb",
        "swh:1:cnt:63d8dbd40c23542e740659a7168a0ce3138ea748",
        "swh:1:cnt:64845fb7679efbd24296bf0499f7386cde449f0a",
        "swh:1:cnt:675f43ab433f9bacd8574d633afe18364af6a107",
        "swh:1:cnt:67c3297611451bb352bbda488b0cd7cb2528a0d3",
        "swh:1:cnt:680fc5115355e433bf39e51c98f66ddb8e3dce22",
        "swh:1:cnt:6b10f9584314d9e7304b2c13c480fc3d01acabe9",
        "swh:1:cnt:6b2aaa7640726588bcd3d57e1de4b1315b7f315e",
        "swh:1:cnt:6bf0c97a7f84620a0bb4cf6380ec307748e043bd",
        "swh:1:cnt:6cc8370785d2e6c51b9de5ff36cda89a1a9aa2f9",
        "swh:1:cnt:6d0b7ebde95e1fbd7cc77a60d46a3468f61038cc",
        "swh:1:cnt:6d7ce64fb8c0411b6a006095186464c3a70de1be",
        "swh:1:cnt:6f4f765ed6998a016a32bf9b4c48b77219ef7f0a",
        "swh:1:cnt:7136b4c89bbca73c68a2ee0eaca246942d6fa2bb",
        "swh:1:cnt:7371f47a6f8bd23a8fa1a8b2a9479cdd76380e54",
        "swh:1:cnt:74e0f12e3246e5d0b556558359a30e0991092cdc",
        "swh:1:cnt:7813681f5b41c028345ca62a2be376bae70b7f61",
        "swh:1:cnt:7937c68fbcf7c484f2d5ce7801944416eedf0d2c",
        "swh:1:cnt:79b7e4bad2a1676d0db76503825455527b89d932",
        "swh:1:cnt:7b71c6e679738d980d5d03559d35fd88a66ee9e4",
        "swh:1:cnt:7c42345690273ab86713294ad761386871c7b708",
        "swh:1:cnt:7de85f863158583176767468e52c7cbb92996542",
        "swh:1:cnt:81750b96f9d83b395f285233d54ec0c9df9ab93d",
        "swh:1:cnt:8214d0ee079917c29e57d16e764fc46de8fb50bf",
        "swh:1:cnt:825026bf2f91b30f020cdba8326485cc600d8a4d",
        "swh:1:cnt:835a5816354544643aeb095a3de6ae8c43136c6d",
        "swh:1:cnt:84f8fe2f54c9b5bf32ddbea402b6ae4af382fa76",
        "swh:1:cnt:851c75cc5e74cf97d145360755d116d2409881f2",
        "swh:1:cnt:866ad4749eb52020cf6b9791205db4af53d337e1",
        "swh:1:cnt:883ad6e8ef9a7392b45f6fc9e7d53c88f502388b",
        "swh:1:cnt:8a908eccd273b758a246975d2eaf96968ceb330d",
        "swh:1:cnt:8ac2eb508956928faf94cbca788daa7b2aeb7735",
        "swh:1:cnt:8b0e2fb7fa3d187b31a6970199816db30d79e8df",
        "swh:1:cnt:8b1296cad20a66969613992eb9e0eb493a1b21fa",
        "swh:1:cnt:8b137891791fe96927ad78e64b0aad7bded08bdc",
        "swh:1:cnt:8b43ca9ac41e863b1817ba9c2631c33ed0b95265",
        "swh:1:cnt:8c53a7f5f36f2172d12e3142124b3da1ee9321b6",
        "swh:1:cnt:8c7e5a667f1b771847fe88c01c3de34413a1b220",
        "swh:1:cnt:8cbf1afc66d01c03210c74c6c3837cf879484525",
        "swh:1:cnt:8e2f0bef135ba8e52e4110b6a5b0ebf19a528ca4",
        "swh:1:cnt:8f7462996850b690a4a2b2d5c368475744ec29fb",
        "swh:1:cnt:9280c0d31d5a7c1fd9abaaa7ffd6160759eca320",
        "swh:1:cnt:92a39f398b80310c58aa2574dd97a3e35527b73b",
        "swh:1:cnt:945c9b46d684f08ec84cb316e1dc0061e361f794",
        "swh:1:cnt:96583aabf5a13fdc518899efd0c92c41da28eb2f",
        "swh:1:cnt:96d80cd6c4e7158dbebd0849f4fb7ce513e5828c",
        "swh:1:cnt:9b26e9b102ab2917db3dc1f6ced91065a38205a8",
        "swh:1:cnt:9beacf8e6f5cd023ac57d0b97432c7cc250c6b96",
        "swh:1:cnt:9c95a6ba6af7377402d44dc60243769a4c604364",
        "swh:1:cnt:9cbe6ea56f225388ae614c419249bfc6d734cc30",
        "swh:1:cnt:9d3cd6889873b86c6796ac62f33b7c8d57d44bc0",
        "swh:1:cnt:9d68933c44f13985b9eb19159da6eb3ff0e574bf",
        "swh:1:cnt:9de294132dec5e172c3245f7453b48ddfd1b2006",
        "swh:1:cnt:9e99dd5ec98fdb081b58e76344a05122e25dcb94",
        "swh:1:cnt:9fb75b8d4f4c7faa7ba59d138746231ada07c7b0",
        "swh:1:cnt:9fc8861321f106d71a197c0e06655b10242f82a1",
        "swh:1:cnt:a2344cdfa5b899d869dde30985646273655014d0",
        "swh:1:cnt:a3871d4508259dc2e3eae0be6fc69d1d3daf0a35",
        "swh:1:cnt:a4a063a1573238b224841eb6556061a7ec1f822e",
        "swh:1:cnt:a4ceb359eba9f667d23d07784d7208f8188fcbbe",
        "swh:1:cnt:a5a3498915908e0e03f66cd28e52b9fc87d1c3d5",
        "swh:1:cnt:a8e73b1b7c227edf8c112fc273e9929b88301b13",
        "swh:1:cnt:aa91d5c03e2002a1057f8075f67a189aeaae3f2b",
        "swh:1:cnt:ac044e5e4649cd149e3d0cf9d23720d299288a1e",
        "swh:1:cnt:ad2823b48f78a0667817300ddcac54c2f6c385e9",
        "swh:1:cnt:ae9780bc629ea3cb16fadbba91e393318af71465",
        "swh:1:cnt:afb09b4e6d0e92bd13951b4f9d3ded90c8c4e89a",
        "swh:1:cnt:b0b2b1c8ddb2dd9e025526d2ea88d82854de6be8",
        "swh:1:cnt:b15531f5dd7f917929f0331f7f65018f9c7c25cc",
        "swh:1:cnt:b1d81e79354c904c6cd8bf5bf91f4c9c9502f381",
        "swh:1:cnt:b4158c40d2962b0ec90bde1019290ae984adf2a3",
        "swh:1:cnt:b516b2c489f1fefd53b4a1905bf37480a8eb056d",
        "swh:1:cnt:b54b8e79bb8d039944e1adf71c97c3c74676ab73",
        "swh:1:cnt:b59c5945f9f68d5e239c35ed33100dd24b27b800",
        "swh:1:cnt:b66e5558e5a67ea386d6f6fea028555b8009d0f9",
        "swh:1:cnt:b7d5379f9e3bb741760e46cf87c89b081f8f44d9",
        "swh:1:cnt:b8635bcf0d86ea51ce65da503c779e72262c0c9e",
        "swh:1:cnt:b9798c2427e725d88f22a1fe75cd165046f614f7",
        "swh:1:cnt:baf72b1da3ee845c0543fe0acf4e02e1a031f397",
        "swh:1:cnt:bb79ec2de59197fe11eeb60d312673a87c1b8932",
        "swh:1:cnt:bb7d13c5e9acc98633147acb5e2e5cd2dda09978",
        "swh:1:cnt:bd0fd35942234660f0c30ba251abaa81eda68fd1",
        "swh:1:cnt:bd97d75340e460bdddedbc17115518f47aa5ba93",
        "swh:1:cnt:be54354a9433a1e798cf17a5cddffbf581e3afa2",
        "swh:1:cnt:bf0d87ab1b2b0ec1a11a3973d2845b42413d9767",
        "swh:1:cnt:c137216fe167556782049b618443432e9409fa53",
        "swh:1:cnt:c1b0730e0133447badcfd47fd144e254807b06e1",
        "swh:1:cnt:c227083464fb9af8955c90d2924774ee50abb547",
        "swh:1:cnt:c2fb4f3370c4e65a8a889f93628d0dd0d0d02726",
        "swh:1:cnt:c30d0581bfa3e1d73eb6a5efbe638c72911dadb7",
        "swh:1:cnt:c32b65a02039b3e986f3b49bdd19f90ffd6e23ff",
        "swh:1:cnt:c471733217fd3bdca8cae8a713b434a3d16ea303",
        "swh:1:cnt:c59d9b6344f1af00e504ba698129f07a34bbed8d",
        "swh:1:cnt:c5fa78456dbde5c0ffa05377925131501c0c794d",
        "swh:1:cnt:c7930257dfef505fd996e1d6f22f2f35149990d0",
        "swh:1:cnt:c96ab3cc70e72f99ab51e3bdab9465fe177c7923",
        "swh:1:cnt:c9cdc63b07017d57efbf45789ceaecc069064a8b",
        "swh:1:cnt:caad5337178380d33d72638694c823b03963b7da",
        "swh:1:cnt:cd571f464ed47d4549f59166841c6a041921342b",
        "swh:1:cnt:ce542efaa5124a0437f0c4db329d7ec4b7ba70a7",
        "swh:1:cnt:d133603777a89a23bce43ec32130942a3c93ba5c",
        "swh:1:cnt:d426ba8a32b9ab2d8f994b193c0302ea46d5e5c3",
        "swh:1:cnt:d50394efdf1c93a17379ff5799d1922b2d436026",
        "swh:1:cnt:d6c11f4cd0a101d19c54e9c378526bdf21988732",
        "swh:1:cnt:d77740a55759db09434d496f10527e0560ee590d",
        "swh:1:cnt:d8263ee9860594d2806b0dfd1bfd17528b0ba2a4",
        "swh:1:cnt:dd67376917e5edd850e1db02049d5c6115462d8d",
        "swh:1:cnt:ddf3b06abfe0500f89028d1b1304c57169da8c87",
        "swh:1:cnt:df3d21358f562c835826070e8ef31e0de97e13a4",
        "swh:1:cnt:e0074173a81d3825f14fe83d3aefc03d51476f00",
        "swh:1:cnt:e0aa8a9ce7243aa018b70b3839d567b14fcf01e0",
        "swh:1:cnt:e25f1814e51579d5f55c0f1fe0135ddb28a47f4a",
        "swh:1:cnt:e2cd95fdbf767aa5e9a1f1bd67167db4a0725d5d",
        "swh:1:cnt:e440e5c842586965a7fb77deda2eca68612b1f53",
        "swh:1:cnt:e49435269d332947aed299a315268e347fb29e7e",
        "swh:1:cnt:e515dc1cf2373da83cc4b248c392c241e4ea2afc",
        "swh:1:cnt:e5db411da00fcba56fe620bad5192fc6a7b1325b",
        "swh:1:cnt:e7754cae5adecf1f21102527fbdeae39280f8e24",
        "swh:1:cnt:e7a5832a436197d3bdc6c6d879a732c627497b0d",
        "swh:1:cnt:e8a0f87653d8b78789cb183ba19f357c636ad33f",
        "swh:1:cnt:ea0c8a85cb7293feae2c9e151d1d395be59b61fa",
        "swh:1:cnt:ea9ed43d2b2445bf0ef9bc4e146f93f329d90109",
        "swh:1:cnt:ec5fd4221a9dbb645852c59f2a455671b4f2bfd0",
        "swh:1:cnt:eda5949cbef3c59679d61a78aa0654d43c3f06fa",
        "swh:1:cnt:eea1bf0c31f3d493509fa1ddd9d25c734eb35277",
        "swh:1:cnt:eec1a4233f3f92568aa3ede33cb1f4ddb6dcc251",
        "swh:1:cnt:ef073cc45ccc66c921adc7ccc6221e38aa54ae17",
        "swh:1:cnt:ef6080906700f3f3cdac7d60341a5de7b5da5581",
        "swh:1:cnt:ef6bce1d1d15c6721aa1c5cce64b10378dfcc844",
        "swh:1:cnt:f11c82a4cb6cc2e8f3bdf52b5cdeaad4d5bb214e",
        "swh:1:cnt:f1b38290cdf3564ec0d9a35bea05c5cb5cd5f503",
        "swh:1:cnt:f327b9acc7ffd3c3008a5363dad25358ca32f643",
        "swh:1:cnt:f3c6c3c68af484bd95bc7b7a38276c6c014d6aa3",
        "swh:1:cnt:f46d387bf94c89b81aaec9cff2ee52d1f9ab4187",
        "swh:1:cnt:f4c82dc4e83cb89b72d1713652f6cdc19ccb2aa4",
        "swh:1:cnt:f59ec20aabf5842d237244ece8c81ab184faeac1",
        "swh:1:cnt:f76dd238ade08917e6712764a16a22005a50573d",
        "swh:1:cnt:f7a8cadeb57c7e92cda09b565e88c21487119d8e",
        "swh:1:cnt:f7cb26f07be14f68a35431ff8ab12b69b61424a3",
        "swh:1:cnt:f8fa5a235408d5c6f2c2844e44cd98c1fa666ac8",
        "swh:1:cnt:f9825866e00f2b8cd33a9dabe1a23a9ee7b3c984",
        "swh:1:cnt:fa7af8bf5fdd704f73beb3adc5612682a98e1af5",
        "swh:1:cnt:fc2b5693e00bdbea3f74ac60459b1df43c295356",
        "swh:1:cnt:fe1990c57a840e97c1075904f9e96daed2e65030",
        "swh:1:cnt:ff30235f076b3fde71fba6ceafa0f510e0006261",
        "swh:1:cnt:ff7dbb100623d2c1f447485012e8de229821cc84",
        # MIT License
        "swh:1:cnt:654d0bfe943437d43242325b1fbcff5f400d84ee",
        "swh:1:cnt:e7af2f77107d73046421ef56c4684cbfdd3c1e89",
        "swh:1:cnt:ee27ba4b4412b0e4a05af5e3d8a005bc6681fdf3",
        # GPLv2
        "swh:1:cnt:d159169d1050894d3ea3b98e1c965c4058208fe1",
        "swh:1:cnt:d511905c1647a1e311e8b20d5930a37a9c2531cd",
        "swh:1:cnt:d60c31a97a544b53039088d14fe9114583c0efc3",
        # GPLv3
        "swh:1:cnt:65c5ca88a67c30becee01c5a8816d964b03862f9",
        "swh:1:cnt:94a9ed024d3859793618152ea559a168bbcbb5e2",
        "swh:1:cnt:9cecc1d4669ee8af2ca727a5d8cde10cd8b2d7cc",
        "swh:1:cnt:f288702d2fa16d3cdf0035b15a9fcbc552cd88e7",
        # AGPLv3
        "swh:1:cnt:dba13ed2ddf783ee8118c6a581dbf75305f816a3",
        # Apache License 2.0
        "swh:1:cnt:261eeb9e9f8b2b4b0d119366dda99c6fd7d35c64",
        "swh:1:cnt:8dada3edaf50dbc082c9a125058f25def75e625a",
        "swh:1:cnt:d645695673349e3947e8e5ae42332d0ac3164cd7",
        # ISC License
        "swh:1:cnt:19129e315fe593965a2fdd50ec0d1253bcbd2ece",
        # Create React App .gitignore
        "swh:1:cnt:4d29575de80483b005c29bfcac5061cd2f45313e",
        # React logo
        "swh:1:cnt:6b60c1042f58d9fabb75485aa3624dddcf633b5c",
        # VisualStudio default .gitattributes
        "swh:1:cnt:1ff0c423042b46cb1d617b81efb715defbe8054d",
        # Very common .gitattributes
        "swh:1:cnt:dfe0770424b2a19faf507a501ebfc23be8f54e7b",
        # Empty directory
        "swh:1:dir:4b825dc642cb6eb9a060e54bf8d69288fbee4904",
        # Directory with empty .keep
        "swh:1:dir:29a422c19251aeaeb907175e9b3219a9bed6c616",
        # Directory with empty .gitkeep
        "swh:1:dir:d564d0bc3dd917926892c55e3706cc116d5b165e",
        # Directory with empty .gitignore
        "swh:1:dir:82e3a754b6a0fcb238b03c0e47d05219fbf9cf89",
        # Directory with empty __init__.py
        "swh:1:dir:9d1dcfdaf1a6857c5f83dc27019c7600e1ffaff8",
        # Empty snapshot
        "swh:1:snp:1a8893e6a86f444e8be8e7bda6cb34fb1735a00e",
    ]
}


class MarkingState(Enum):
    """Represents the different state of vertices while the marking algorithm is on-going.

    All vertices start UNMARKED. After examination of their inbound references, they are
    either marked as REMOVABLE or UNREMOVABLE.

    :meta private:
    """

    UNMARKED = auto()
    REMOVABLE = auto()
    UNREMOVABLE = auto()


class RemovableSubgraph(Subgraph):
    """A class representing the subgraph of objects that can be safely removed
    from the archive

    This subgraph should only be created from an :py:class:`InventorySubgraph`
    using :py:meth:`from_inventory_subgraph`. No new vertices or edges can be
    added at this point.

    Getting the list of SWHID that can be removed is done by using
    py:meth:`removable_swhids`.

    After marking happened, this subgraph will contain all candidates until
    :py:meth:`delete_unremovable` is called.
    """

    def __init__(self, *args, **kwargs):
        """ """
        super().__init__(*args, **kwargs)
        self["name"] = "Removable"
        self.vs["state"] = MarkingState.UNMARKED

    @classmethod
    def from_inventory_subgraph(cls, subgraph: InventorySubgraph):
        return cls.copy(subgraph)

    def add_vertex(self, name: str, **kwargs):
        """Not available for RemovableSubgraph.

        :meta private:"""
        raise NotImplementedError("No new vertex should be added at this stage.")

    def add_edge(self, src: Vertex, dst: Vertex, skip_duplicates=False, **kwargs):
        """Not available for RemovableSubgraph.

        :meta private:"""
        raise NotImplementedError("No new edge should be added at this stage.")

    def delete_unremovable(self):
        """Delete all vertices for unremovable objects from this subgraph."""

        self.delete_vertices(self.vs.select(state_eq=MarkingState.UNREMOVABLE))

    def removable_swhids(self) -> List[ExtendedSWHID]:
        """Returns a list of SWHIDs that can safely be removed from the archive."""

        return [
            v["swhid"] for v in self.select_ordered(state_eq=MarkingState.REMOVABLE)
        ]

    def referenced_swhids(self) -> List[ExtendedSWHID]:
        """Returns a list of SWHIDs that are referenced by objects that can be
        safely removed from the archive."""

        referenced: Set[Vertex] = set()
        for removable in self.select_ordered(state_eq=MarkingState.REMOVABLE):
            referenced.update(
                v
                for v in removable.successors()
                if v["state"] == MarkingState.UNREMOVABLE
            )
        return [v["swhid"] for v in referenced]

    def dot_node_attributes(self, v: Vertex) -> List[str]:
        """Get a list of attributes in DOT format for the given vertex.

        On top of default attributes, color the background of each node if
        they can be removed or not.

        :meta private:
        """
        attrs = super().dot_node_attributes(v)
        # Unset the shape so we can better spot ExtendedSWHID that can be removed
        if v["state"] == MarkingState.UNMARKED:
            attrs.append('style="filled,dotted"')
        elif v["state"] == MarkingState.UNREMOVABLE:
            attrs.append("fillcolor=white")
        else:
            attrs.append("fillcolor=red")
        return attrs


class Marker:
    """A class encapsulating our algorithm marking nodes as removable or not.

    :meta private:
    """

    def __init__(
        self,
        storage: StorageInterface,
        graph_client: RemoteGraphClient,
        subgraph: RemovableSubgraph,
        progressbar: ProgressBar[Vertex],
    ):
        self._storage = storage
        self._graph_client = graph_client
        self._subgraph = subgraph
        self._progressbar = progressbar

    @property
    def subgraph(self):
        return self._subgraph

    def has_unknown_inbound_edges(self, vertex: Vertex):
        # Origins never have any inbound edges
        if vertex["swhid"].object_type == ObjectType.ORIGIN:
            return False
        known_predecessors = {pred["swhid"] for pred in vertex.predecessors()}
        # We only need to find one extra edge from the ones we already know
        search_limit = len(known_predecessors) + 1
        for pred in chain(
            self.inbound_edges_from_graph(vertex, search_limit),
            self.inbound_edges_from_storage(vertex, search_limit),
        ):
            # Removing a revision used as a submodule should be possible.
            # (We do not care that it will make submodule users miss part
            # of their source code.)
            # We use string matching here because the graph API returns
            # str objects and not ExtendedSWHID objects.
            if (
                vertex["swhid"].object_type == ObjectType.REVISION
                and pred.object_type == ObjectType.DIRECTORY
            ):
                logger.debug(
                    "Skipping predecessor %s of %s as it's a submodule",
                    pred,
                    vertex["swhid"],
                )
                return False
            if pred not in known_predecessors:
                logger.debug(
                    "%s has an extra predecessor (%s), skipping removal",
                    vertex["swhid"],
                    pred,
                )
                return True
        return False

    def inbound_edges_from_graph(
        self, vertex: Vertex, search_limit: int
    ) -> Iterator[ExtendedSWHID]:
        # We need to check if the object still exists in storage, as it
        # might have been removed since the graph was exported.
        try:

            def get_objects(limit: int) -> List[ExtendedSWHID]:
                return [
                    ExtendedSWHID.from_string(str)
                    for str in self._graph_client.neighbors(
                        str(vertex["swhid"]),
                        direction="backward",
                        max_matching_nodes=search_limit,
                    )
                ]

            yield from get_filtered_objects(self._storage, get_objects, search_limit)
        except GraphArgumentException:
            yield from ()

    def inbound_edges_from_storage(
        self, vertex: Vertex, search_limit: int
    ) -> Iterator[ExtendedSWHID]:
        # We have to filter objects missing that can be returned by
        # `object_find_recent_references()`. As we don’t clean up
        # the `object_references` table, this happens if we remove
        # an object in a first pass, before trying to remove one
        # of its targets at a later time.
        yield from get_filtered_objects(
            self._storage,
            partial(self._storage.object_find_recent_references, vertex["swhid"]),
            search_limit,
        )

    def mark_known_missing(self, known_missing: Set[ExtendedSWHID]):
        for swhid in known_missing:
            for v in self._subgraph.vs.select(swhid_eq=swhid):
                v["state"] = MarkingState.UNREMOVABLE

    def mark_candidates(self):
        for index, vid in enumerate(self._subgraph.topological_sorting()):
            vertex = self._subgraph.vs[vid]
            assert all(
                pred["state"] != MarkingState.UNMARKED for pred in vertex.predecessors()
            ), "topological sort broken: one predecessor is still in unmarked state"

            self._progressbar.update(1)

            if vertex["swhid"] in NEVER_REMOVABLE:
                vertex["state"] = MarkingState.UNREMOVABLE
                continue

            # If any predecessor is unremovable, this makes this object unremovable too
            if any(
                pred["state"] == MarkingState.UNREMOVABLE
                for pred in vertex.predecessors()
            ):
                vertex["state"] = MarkingState.UNREMOVABLE
                continue
            vertex["state"] = (
                MarkingState.UNREMOVABLE
                if self.has_unknown_inbound_edges(vertex)
                else MarkingState.REMOVABLE
            )


def mark_removable(
    storage: StorageInterface,
    graph_client: RemoteGraphClient,
    inventory_subgraph: InventorySubgraph,
    known_missing: Optional[Set[ExtendedSWHID]] = None,
    progressbar: Optional[ProgressBarInit] = None,
) -> RemovableSubgraph:
    """Find which candidates can be safely removed from the archive.

    Using the previously populated ``inventory_subgraph``, query the given
    ``storage`` and ``graph_client`` to assess inbound references for each
    candidate and see which one can be safely removed from the archive.
    """
    subgraph = RemovableSubgraph.from_inventory_subgraph(inventory_subgraph)
    progressbar_init: ProgressBarInit = progressbar or no_progressbar
    bar: ProgressBar[int]
    with progressbar_init(
        length=len(subgraph.vs),
        label="Determining which objects can be safely removed…",
    ) as bar:
        marker = Marker(storage, graph_client, subgraph, bar)
        marker.mark_candidates()
        if known_missing:
            marker.mark_known_missing(known_missing)
    return marker.subgraph



# Copyright (C) 2023 The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from __future__ import annotations

import logging
import pathlib
import sys
from typing import (
    TYPE_CHECKING,
    Callable,
    Dict,
    Iterable,
    Optional,
    Set,
    TextIO,
    Tuple,
    cast,
)

import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group

if TYPE_CHECKING:
    from swh.model.model import Origin
    from swh.model.swhids import ExtendedSWHID

    from .operations import Remover
    from .progressbar import ProgressBar, V
    from .recovery_bundle import ObjectDecryptionKeyProvider, ShareDecryptionKeys


class SWHIDType(click.ParamType):
    name = "swhid"

    def convert(self, value, param, ctx) -> "ExtendedSWHID":
        from swh.model.swhids import ExtendedSWHID, ValidationError

        try:
            return ExtendedSWHID.from_string(value)
        except ValidationError:
            raise click.ClickException(f"Unable to parse “{value}” as a SWHID.")


class SwhidOrUrlParamType(click.ParamType):
    name = "swhid or origin URL"

    def convert(self, value, param, ctx):
        from swh.model.exceptions import ValidationError
        from swh.model.model import Origin
        from swh.model.swhids import ExtendedSWHID

        if value.startswith("swh:1:"):
            try:
                return ExtendedSWHID.from_string(value)
            except ValidationError:
                self.fail(f"expected extended SWHID, got {value!r}", param, ctx)
        else:
            return Origin(url=value)


class ClickLoggingHandler(logging.Handler):
    """Handler displaying logs using click.secho(), passing the style extra
    attribute."""

    def emit(self, record):
        if hasattr(record, "style"):
            click.secho(self.format(record), **record.style)
        else:
            click.echo(self.format(record))


def progressbar(
    iterable: Optional[Iterable[V]] = None,
    length: Optional[int] = None,
    label: Optional[str] = None,
    show_eta: bool = True,
    show_pos: bool = False,
    show_percent: Optional[bool] = None,
    item_show_func: Optional[Callable[[V], str]] = None,
) -> ProgressBar[V]:
    bar = click.progressbar(
        iterable=iterable,
        length=length,
        label=label,
        show_eta=show_eta,
        show_pos=show_pos,
        show_percent=show_percent,
        item_show_func=item_show_func,
        file=sys.stderr,
    )
    # We have to use `cast()` to renconcile the case where
    # length is used and `click.progressbar()` returns a
    # `ProgressBar[int]`. But in that case, iterable is not
    # given, so V is not bound and it is safe to assume
    # that V = int.
    return cast("ProgressBar[V]", bar)


@swh_cli_group.group(name="alter", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def alter_cli_group(ctx):
    """Archive alteration tools.

    Location of the configuration should be specified through the environment
    variable ``SWH_CONFIG_FILENAME``.

    Expected config format:

        \b
        storage:
          cls: remote
          url: https://storage-cassandra-ro.softwareheritage.org
        \b
        graph:
          url: "http://granet.internal.softwareheritage.org:5009/graph"
        \b
        restoration_storage:
          cls: remote
          url: https://storage-rw.softwareheritage.org
        \b
        removal_searches:
          main:
            cls: elasticsearch
            hosts:
            - elasticsearch:9200
        \b
        removal_storages:
          old_primary:
            cls: postgresql
            db: "service=swh"
          new_primary:
            cls: cassandra
            hosts:
            - cassandra-seed
            keyspace: swh
        \b
        removal_objstorages:
          main:
            cls: remote
            url: https://objstorage.softwareheritage.org
        \b
        removal_journals:
          main_journal:
            cls: kafka
            brokers:
            - kafka1.internal.softwareheritage.org
            prefix: swh.journal.objects
            client_id: swh.alter.removals
        \b
        recovery_bundles:
          secret_sharing:
            minimum_required_groups: 2
            groups:
              legal:
                minimum_required_shares: 1
                recipient_keys:
                    "YubiKey serial 4245067 slot 1": age1yubikey1q2e37f74zzazz75mtggzql3at66pegemfnul0dtd7axctahljkvsqezscaq
                    "YubiKey serial 2284622 slot 3": age1yubikey4o1aypv83isatti92q1zasv1hkpuozlkoak4zd66t7poud23rftqrcszjgul
              sysadmins:
                minimum_required_shares: 1
                recipient_keys:
                    "YubiKey serial 3862152 slot 1": age1yubikeyrupnxsu6uneqxw146g9szaofyxexiy4nhnzqg1ayb9b85g8h4oardwj6c212
                    "Ruby": age1y6epp27nq8n4faj8g8hkw8thcvj744y5vnr8jyfmp4857d6npc3qn9k7jz
        \b
        journal_writer:
          cls: kafka
          brokers:
          - kafka1.internal.softwareheritage.org
          prefix: swh.journal
          client_id: swh.alter.removals

    The identifier for the recipient key must be in the form of
    “YubiKey serial ####### slot #” if the secret key is stored
    on a YubiKey. Keys specified by any other identifiers will be
    considered as plain age identities.
    """  # noqa: B950
    from swh.core import config

    from .operations import logger as operations_logger
    from .recovery_bundle import logger as recovery_bundle_logger

    try:
        conf = config.load_from_envvar()
    except AssertionError as ex:
        raise click.ClickException(ex.args[0])
    ctx.ensure_object(dict)
    ctx.obj["config"] = conf

    for logger in (operations_logger, recovery_bundle_logger):
        if not logger.propagate:
            # Avoid configuring the logger twice
            continue
        logger.propagate = False
        logger.addHandler(ClickLoggingHandler())

    return ctx


def read_swhids(file: TextIO) -> Set["ExtendedSWHID"]:
    import re

    from swh.model.swhids import ExtendedSWHID

    filter_re = re.compile(r"^(#|$)")
    return {
        ExtendedSWHID.from_string(line.strip())
        for line in file.read().split("\n")
        if not filter_re.match(line)
    }


def get_remover(
    ctx: click.Context,
    dry_run: bool = False,
    require_masking_admin: bool = False,
    ignore_backends: Optional[Iterable[str]] = None,
) -> "Remover":
    from psycopg2 import OperationalError, ProgrammingError

    from swh.core.api import RemoteException
    from swh.graph.http_client import GraphAPIError, RemoteGraphClient
    from swh.journal.writer import get_journal_writer
    from swh.journal.writer.kafka import KafkaJournalWriter
    from swh.objstorage.factory import get_objstorage
    from swh.objstorage.interface import ObjStorageInterface
    from swh.search import get_search
    from swh.search.interface import SearchInterface
    from swh.storage import get_storage
    from swh.storage.interface import ObjectDeletionInterface

    from .operations import Remover

    conf = ctx.obj["config"]
    if ignore_backends is None:
        ignore_backends = []

    if "graph" not in conf:
        if "graph" in ignore_backends:
            graph_client = None
        else:
            raise click.ClickException("Configuration does not define `graph`")
    else:
        try:
            graph_client = RemoteGraphClient(**conf["graph"])
        except GraphAPIError as e:
            raise click.ClickException(
                f"Unable to connect to the graph server: {e.args}"
            )

    storage = get_storage(**conf["storage"])

    if not dry_run:
        if "restoration_storage" not in conf:
            raise click.ClickException(
                "Configuration does not define `restoration_storage`"
            )
        if "search" not in ignore_backends and (
            "removal_searches" not in conf or len(conf["removal_searches"]) == 0
        ):
            raise click.ClickException(
                "Configuration does not define any `removal_searches`"
            )
        if "removal_storages" not in conf or len(conf["removal_storages"]) == 0:
            raise click.ClickException(
                "Configuration does not define any `removal_storages`"
            )
        if "removal_objstorages" not in conf or len(conf["removal_objstorages"]) == 0:
            raise click.ClickException(
                "Configuration does not define any `removal_objstorages`"
            )
        if "journal" not in ignore_backends and (
            "removal_journals" not in conf or len(conf["removal_journals"]) == 0
        ):
            raise click.ClickException(
                "Configuration does not define any `removal_journals`"
            )

    restoration_storage = (
        get_storage(**conf["restoration_storage"])
        if "restoration_storage" in conf
        else None
    )

    removal_searches = {}
    for name, d in conf.get("removal_searches", {}).items():
        removal_searches[name] = get_search(**d)
        try:
            removal_searches[name].check()
        except RemoteException as e:
            raise click.ClickException(f"Search “{name}” is unreachable: {e}")

    removal_storages = {}
    for name, d in conf.get("removal_storages", {}).items():
        removal_storage = get_storage(**d)
        assert hasattr(
            removal_storage, "object_delete"
        ), f"storage “{name}” does not implement ObjectDeletionInterface"
        removal_storages[name] = removal_storage

    removal_objstorages = {}
    for name, d in conf.get("removal_objstorages", {}).items():
        removal_objstorages[name] = get_objstorage(**d)

    removal_journals = {}
    for name, d in conf.get("removal_journals", {}).items():
        journal_writer = get_journal_writer(**d)
        assert isinstance(
            journal_writer, KafkaJournalWriter
        ), "journal writer is not kafka-based"
        removal_journals[name] = journal_writer

    known_missing = set(ctx.params.get("known_missing_swhids", set()))
    if known_missing_file := ctx.params.get("known_missing_file"):
        known_missing.update(read_swhids(known_missing_file))

    if require_masking_admin:
        from swh.storage.proxies.masking.db import MaskingAdmin

        if "masking_admin" not in conf or "db" not in conf["masking_admin"]:
            raise click.ClickException("masking_admin.db not found in configuration")
        try:
            masking_admin = MaskingAdmin.connect(conf["masking_admin"]["db"])
        except (OperationalError, ProgrammingError) as e:
            raise click.ClickException(f"Unable to connect to masking database: {e}")
    else:
        masking_admin = None

    return Remover(
        storage=storage,
        graph_client=graph_client,
        restoration_storage=restoration_storage,
        removal_searches=cast(Dict[str, SearchInterface], removal_searches),
        removal_storages=cast(Dict[str, ObjectDeletionInterface], removal_storages),
        removal_objstorages=cast(Dict[str, ObjStorageInterface], removal_objstorages),
        removal_journals=cast(Dict[str, KafkaJournalWriter], removal_journals),
        masking_admin=masking_admin,
        known_missing=known_missing,
        progressbar=progressbar,
    )


@alter_cli_group.command()
@click.option(
    "--dry-run",
    type=click.Choice(
        ["stop-before-recovery-bundle", "stop-before-removal"], case_sensitive=False
    ),
    help="perform a trial run",
)
@click.option(
    "--output-inventory-subgraph",
    type=click.File(mode="w", atomic=True),
)
@click.option(
    "--output-removable-subgraph",
    type=click.File(mode="w", atomic=True),
)
@click.option(
    "--output-pruned-removable-subgraph",
    type=click.File(mode="w", atomic=True),
)
@click.option(
    "--identifier",
    metavar="IDENTIFIER",
    required=True,
    help="identifier for this removal operation",
)
@click.option(
    "--reason",
    metavar="REASON",
    help="reason for this removal operation",
)
@click.option(
    "--expire",
    metavar="YYYY-MM-DD",
    type=click.DateTime(formats=["%Y-%m-%d"]),
    help="date when the recovery bundle should be removed",
)
@click.option(
    "--recovery-bundle",
    metavar="PATH",
    type=click.Path(dir_okay=False),
    required=True,
    help="path to the recovery bundle that will be created",
)
@click.option(
    "--known-missing",
    "known_missing_swhids",
    metavar="SWHID",
    type=SWHIDType(),
    multiple=True,
    help="object known to be missing from storage",
)
@click.option(
    "--known-missing-file",
    "known_missing_file",
    metavar="PATH",
    type=click.File(),
    help=(
        "file (or '-') with object known to be missing from storage, "
        "one SWHID per line"
    ),
)
@click.option(
    "--allow-empty-content-objects/--disallow-empty-content-objects",
    "allow_empty_content_objects",
    default=False,
    help="Create recovery bundle even when data for Content object cannot be found",
)
@click.option(
    "--ignore",
    "ignore_backends",
    default=None,
    multiple=True,
    type=click.Choice(["search", "journal", "graph"], case_sensitive=False),
    help=(
        "Do not make the given backend mandatory when checking the configuration; "
        "this command is usually meant to remove objects from all possible data silos, "
        "so the default behavior is to have all of them mandatory in the configuration "
        "file. Using this option allows to explicitly ignore the given 'backend'"
    ),
)
@click.argument(
    "requested",
    metavar="<SWHID|URL>..",
    type=SwhidOrUrlParamType(),
    required=True,
    nargs=-1,
)
@click.pass_context
def remove(
    ctx,
    requested: Tuple["Origin" | "ExtendedSWHID", ...],
    dry_run: bool,
    output_inventory_subgraph,
    output_removable_subgraph,
    output_pruned_removable_subgraph,
    identifier,
    reason,
    expire,
    recovery_bundle,
    known_missing_swhids,
    known_missing_file,
    allow_empty_content_objects,
    ignore_backends,
) -> None:
    """Remove the given SWHIDs or URLs from the archive."""

    from swh.journal.writer import JournalWriterInterface, get_journal_writer
    from swh.model.model import Origin

    from .inventory import RootsNotFound, StuckInventoryException
    from .notifications import RemovalNotification
    from .operations import RemoverError
    from .recovery_bundle import ContentDataNotFound, SecretSharing

    try:
        secret_sharing = SecretSharing.from_dict(
            ctx.obj["config"]["recovery_bundles"]["secret_sharing"]
        )
    except ValueError as e:
        raise click.ClickException(f"Wrong secret sharing configuration: {e.args[0]}")

    if dry_run != "stop-before-recovery-bundle":
        try:
            p = pathlib.Path(recovery_bundle)
            p.touch(exist_ok=False)
            p.unlink()
        except FileExistsError:
            raise click.ClickException(f"File “{recovery_bundle}” already exists")
        except PermissionError:
            raise click.ClickException(f"Permission denied: “{recovery_bundle}”")

    remover = get_remover(ctx, dry_run, ignore_backends=ignore_backends)

    swhids = [x.swhid() if isinstance(x, Origin) else x for x in requested]

    journal_writer: Optional[JournalWriterInterface] = None
    if "journal_writer" in ctx.obj["config"]:
        cfg = ctx.obj["config"]["journal_writer"]
        journal_writer = get_journal_writer(**cfg)

    try:
        removable = remover.get_removable(
            swhids,
            output_inventory_subgraph=output_inventory_subgraph,
            output_removable_subgraph=output_removable_subgraph,
            output_pruned_removable_subgraph=output_pruned_removable_subgraph,
        )
        removable.print_plan()
        if dry_run == "stop-before-recovery-bundle":
            click.echo("Stopping before creating the recovery bundle as requested.")
            ctx.exit(0)

        if dry_run is None:
            click.confirm(
                click.style(
                    "Proceed?",
                    fg="yellow",
                    bold=True,
                ),
                abort=True,
            )

        decryption_key = remover.create_recovery_bundle(
            secret_sharing=secret_sharing,
            requested=list(requested),
            removable=removable,
            recovery_bundle_path=recovery_bundle,
            removal_identifier=identifier,
            reason=reason,
            expire=expire.astimezone() if expire else None,
            allow_empty_content_objects=allow_empty_content_objects,
        )
        click.secho(f"Recovery bundle decryption key: {decryption_key}", fg="blue")
    except RemoverError as e:
        click.secho(e.args[0], err=True, fg="red")
        ctx.exit(1)
    except RootsNotFound as e:
        click.secho(
            "Some requested objects were not found:",
            err=True,
            fg="red",
            bold=True,
        )
        for label in e.get_labels(requested):
            click.secho(f"- {label}", err=True)
        ctx.exit(1)
    except StuckInventoryException as e:
        click.secho(
            "Inventory phase got stuck. "
            "Unable to learn the complete set of what these objects reference:",
            err=True,
            fg="red",
            bold=True,
        )
        click.secho("\n".join(f"- {swhid}" for swhid in e.swhids), err=True, fg="red")
        ctx.exit(1)
    except ContentDataNotFound as e:
        click.secho(
            f"Content “{e.swhid}” exists, but its data was not found.",
            err=True,
            fg="red",
            bold=True,
        )
        click.secho(
            "Consider using `--allow-empty-content-objects` but only "
            "if the above is expected.",
            err=True,
            fg="yellow",
        )
        ctx.exit(1)

    if dry_run == "stop-before-removal":
        click.echo("Stopping before removal.")
        ctx.exit(0)

    try:
        remover.remove()
    except Exception as e:
        click.secho(str(e), err=True, fg="red", bold=True)
        click.secho("Rolling back…", fg="cyan")
        remover.restore_recovery_bundle()
        ctx.exit(1)
    else:
        if journal_writer is not None:
            click.secho("Sending removal notification…", fg="cyan")
            notif = RemovalNotification(
                removal_identifier=identifier,
                reason=reason,
                requested=list(swhids),
                removed_objects=remover.swhids_to_remove,
            )
            journal_writer.write_addition("removal_notification", notif)
            journal_writer.flush()
            click.secho("Removal notification sent.", fg="green")


@alter_cli_group.command("list-candidates")
@click.option(
    "--omit-referenced/--no-omit-referenced",
    default=True,
    help="Omit candidates that are referenced by other objects",
)
@click.argument(
    "requested",
    metavar="<SWHID|URL>..",
    type=SwhidOrUrlParamType(),
    required=True,
    nargs=-1,
)
@click.pass_context
def list_candidates(
    ctx: click.Context,
    requested: Tuple["Origin" | "ExtendedSWHID", ...],
    omit_referenced: bool,
):
    """List candidates for an altering operation (e.g. removal)

    Display a list of SWHIDs of objects that would be affected by an altering
    operation targeting the SWHIDs (or origin URLs) given as arguments.

    Candidates referenced by objects in the graph outside the set of candidates
    will be filtered out, unless `--no-omit-referenced` is given.
    """

    from swh.graph.http_client import GraphAPIError, RemoteGraphClient
    from swh.model.model import Origin
    from swh.storage import get_storage

    from .inventory import (
        RootsNotFound,
        StuckInventoryException,
        get_raw_extrinsic_metadata,
        make_inventory,
    )
    from .removable import mark_removable

    conf = ctx.obj["config"]

    try:
        graph_client = RemoteGraphClient(**conf["graph"])
    except GraphAPIError as e:
        raise click.ClickException(f"Unable to connect to the graph server: {e.args}")

    storage = get_storage(**conf["storage"])

    swhids = [x.swhid() if isinstance(x, Origin) else x for x in requested]
    try:
        subgraph = make_inventory(
            storage, graph_client, swhids, progressbar=progressbar
        )
    except StuckInventoryException as e:
        click.secho(
            "Inventory phase got stuck. "
            "Unable to learn the complete set of what these objects reference:",
            err=True,
            fg="red",
            bold=True,
        )
        click.secho("\n".join(f"- {swhid}" for swhid in e.swhids), err=True, fg="red")
        ctx.exit(1)
    except RootsNotFound as e:
        click.secho(
            "Some requested objects were not found:",
            err=True,
            fg="red",
            bold=True,
        )
        for label in e.get_labels(requested):
            click.secho(f"- {label}", err=True)
        ctx.exit(1)
    if omit_referenced:
        subgraph = mark_removable(
            storage, graph_client, subgraph, progressbar=progressbar
        )
        subgraph.delete_unremovable()
    removable_swhids = list(subgraph.swhids())
    removable_swhids.extend(
        get_raw_extrinsic_metadata(storage, removable_swhids, progressbar=progressbar)
    )
    for swhid in removable_swhids:
        click.echo(swhid)


@alter_cli_group.command("run-mirror-notification-watcher")
@click.pass_context
def run_mirror_notification_watcher(ctx: click.Context):
    """Watch the journal for notifications from the main archive.

    For removal notifications, we mask the associated objects until
    a decision is made by the mirror operators.

    Example configuration:

        \b
        journal_client:
          brokers: kafka.example.org:9092
          group_id: mirror-notification-watcher
          prefix: swh.journal
        storage:
          cls: remote
          url: https://storage-ro
        masking_admin:
          cls: postgresql
          db: service=masking-db-rw
        emails:
          from: swh-mirror@example.org
          recipients:
          - trinity@example.org
          - neo@example.org
        smtp:
          host: localhost
          port: 25

    The addresses listed as “recipients” in the “emails” section will receive
    an email to let them know that a decision needs to be taken.
    """

    from swh.alter.mirror_notification_watcher import MirrorNotificationWatcher
    from swh.journal.client import get_journal_client
    from swh.storage import get_storage
    from swh.storage.proxies.masking.db import MaskingAdmin

    conf = ctx.obj["config"]

    try:
        storage = get_storage(**conf["storage"])
        storage.check_config(check_write=False)
    except Exception as e:
        raise click.ClickException(f"Unable to query to the storage: {e}")

    try:
        journal_client = get_journal_client(
            **{
                **conf["journal_client"],
                "cls": "kafka",
                "object_types": ["removal_notification"],
            }
        )
    except Exception as e:
        raise click.ClickException(f"Unable to setup the journal client: {repr(e)}")

    try:
        masking_admin_dsn = conf["masking_admin"]["db"]
        _ = MaskingAdmin.connect(masking_admin_dsn)
    except Exception as e:
        raise click.ClickException(
            f"Unable to connect to the masking proxy database: {repr(e)}"
        )

    emails_from = conf["emails"].get("from")
    if emails_from is None:
        raise click.ClickException("“emails.from” has not been set.")
    emails_recipients = conf["emails"].get("recipients")
    if emails_recipients is None:
        raise click.ClickException("“emails.recipients” has not been set.")
    if not isinstance(emails_recipients, list) or len(emails_recipients) < 1:
        raise click.ClickException(
            "“emails.recipients” must be a list and contain at least one email address."
        )

    smtp_host = conf["smtp"].get("host")
    if smtp_host is None:
        raise click.ClickException("“smtp.host” has not been set.")
    try:
        smtp_port = int(conf["smtp"].get("port"))
    except Exception as e:
        raise click.ClickException(f"“smtp.port” must be set to a port number: {e}")

    watcher = MirrorNotificationWatcher(
        storage=storage,
        journal_client=journal_client,
        masking_admin_dsn=masking_admin_dsn,
        emails_from=emails_from,
        emails_recipients=emails_recipients,
        smtp_host=smtp_host,
        smtp_port=smtp_port,
    )
    try:
        watcher.watch()
    except KeyboardInterrupt:
        ctx.exit(0)


@alter_cli_group.group(name="recovery-bundle", context_settings=CONTEXT_SETTINGS)
@click.pass_context
def recovery_bundle_cli_group(ctx):
    """Recovery bundle related tools."""
    return ctx


@recovery_bundle_cli_group.command(name="info")
@click.option(
    "--dump-manifest",
    is_flag=True,
    default=False,
    help="Show raw manifest in YAML format.",
)
@click.option(
    "--show-encrypted-secrets",
    is_flag=True,
    default=False,
    help="Show encrypted secrets.",
)
@click.argument(
    "recovery-bundle",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=True,
)
@click.pass_context
def info(ctx, recovery_bundle, dump_manifest, show_encrypted_secrets) -> None:
    """Display the manifest of the given recovery bundle."""
    from swh.model.model import Origin

    from .recovery_bundle import RecoveryBundle

    bundle = RecoveryBundle(recovery_bundle)

    if dump_manifest:
        click.echo(bundle.dump_manifest(), nl=False)
        ctx.exit()

    title = f"Recovery bundle “{bundle.removal_identifier}”"
    click.echo(title)
    click.echo("=" * len(title))
    click.echo("")
    click.echo(f"Created: {bundle.created.isoformat()}")
    if bundle.reason:
        lines = bundle.reason.rstrip().split("\n")
        lines[0] = f"Reason: {lines[0]}"
        click.echo("\n        ".join(lines))
    if bundle.expire:
        click.echo(f"Expire: {bundle.expire}")
    if bundle.version >= 3:
        click.echo("Removal requested for:")
        for x in bundle.requested:
            click.echo(f"- {x.url if isinstance(x, Origin) else x}")
    click.echo("SWHID of the objects present in the bundle:")
    for swhid in bundle.swhids:
        click.echo(f"- {swhid}")
    if bundle.version >= 3 and len(bundle.referencing):
        click.echo("SWHID referenced by objects in this bundle:")
        for swhid in bundle.referencing:
            click.echo(f"- {swhid}")
    click.echo("Secret share holders:")
    for share_id in sorted(bundle.share_ids):
        click.echo(f"- {share_id}")
        if show_encrypted_secrets:
            click.echo(bundle.encrypted_secret(share_id))


def _share_decryption_keys_provider(share_ids: Set[str]) -> ShareDecryptionKeys:
    import subprocess
    import sys

    from .recovery_bundle import list_yubikey_identities

    for attempt in range(1, 10):
        if not any(share_id.startswith("YubiKey") for share_id in share_ids):
            # No shares require a YubiKey, so there is nothing we can do here
            break
        try:
            for share_id, secret_key in list_yubikey_identities():
                if share_id not in share_ids:
                    continue
                share_ids.remove(share_id)
                click.echo(
                    "🔧 Decrypting share using "
                    f"{click.style(share_id, fg='magenta', bold=True)}…"
                )
                click.echo("💭 You might need to tap the right YubiKey when it blinks.")
                yield share_id, secret_key
                click.echo()
        except subprocess.CalledProcessError as ex:
            if "age-plugin-yubikey" not in ex.cmd[0]:
                raise
            click.echo(
                f"""💥 {click.style('age-plugin-yubikey failed to '
                                    'list connected YubiKeys.', bold=True, fg='red')}"""
            )
            click.echo("💭 Please disconnect all YubiKeys and retry.")
            sys.exit(1)
        if share_ids:
            yubikey_ids = list(sorted(share_ids))
            if len(yubikey_ids) > 1:
                yubikeys = ", ".join(
                    click.style(share_id, fg="magenta", bold=True)
                    for share_id in yubikey_ids[:-1]
                )
                yubikeys += " or " + click.style(
                    yubikey_ids[-1], fg="magenta", bold=True
                )
            else:
                yubikeys = click.style(yubikey_ids[0], fg="magenta", bold=True)
            click.prompt(
                f"🔐 Please insert {yubikeys} and press "
                f"{click.style('Enter', fg='green', bold=True)}…",
                default="Ok",
                show_default=False,
                hide_input=True,
                prompt_suffix="",
            )
    click.echo(
        f"""💥 {click.style('Unable to decrypt enough shared secrets to recover '
                                   'the object decryption key. Aborting.',
                                   bold=True, fg='red')}"""
    )
    sys.exit(1)


def _print_decrypted_mnemonic(mnemonic: str, share_id: Optional[str] = None) -> None:
    fmt_from = ""
    if share_id:
        fmt_from = f" from {click.style(share_id, fg='magenta', bold=True)}"
    click.echo(f"🔑 Recovered shared secret{fmt_from}:")
    # Quoting from SLIP-0039: This construction yields a beneficial
    # property where the random identifier and the iteration exponent
    # transform into the first two words of the mnemonic code, so the user
    # can immediately tell whether the correct shares are being combined,
    # i.e. they have to have the same first two words. Moreover, the third
    # word encodes the group index, group threshold and part of the group
    # count. Since the group threshold and group count are constant, all
    # **shares belonging to the same group start with the same three words**.
    words = mnemonic.split()
    click.echo(
        " ".join(
            click.style(word, fg="blue", bold=index < 3)
            for index, word in enumerate(words)
        )
    )


def _recover_mnemonics_from_identity_files(
    manifest, share_ids, identity_files, show_decrypted_mnemonics
):
    from .recovery_bundle import WrongDecryptionKey, age_decrypt_from_identity

    # As we can’t know which identity file corresponds to which encrypted shared
    # secret, we have to try them all and see which one we can actually decrypt.
    recovered = {}
    for identity_file in identity_files:
        for share_id in share_ids:
            try:
                recovered[share_id] = age_decrypt_from_identity(
                    identity_file, manifest.decryption_key_shares[share_id]
                ).decode("us-ascii")
                if show_decrypted_mnemonics:
                    _print_decrypted_mnemonic(recovered[share_id], share_id)
            except WrongDecryptionKey:
                pass
    return recovered


def prompting_object_decryption_key_provider(
    manifest, known_mnemonics=None, identity_files=None, show_decrypted_mnemonics=False
) -> str:
    import functools

    from .recovery_bundle import recover_object_decryption_key_from_encrypted_shares

    decrypted_mnemonic_processor = None
    if show_decrypted_mnemonics:
        decrypted_mnemonic_processor = _print_decrypted_mnemonic
    share_ids = set(manifest.decryption_key_shares.keys())
    # Normalize known_mnemonics
    known_mnemonics = list(known_mnemonics or [])
    if identity_files:
        recovered = _recover_mnemonics_from_identity_files(
            manifest, share_ids, identity_files, show_decrypted_mnemonics
        )
        share_ids.difference_update(recovered.keys())
        known_mnemonics.extend(recovered.values())
    yubikey_share_ids = set(
        share_id for share_id in share_ids if share_id.startswith("YubiKey")
    )
    missing_ids = share_ids - yubikey_share_ids
    if missing_ids:
        fmt_ids = ", ".join(
            click.style(share_id, fg="magenta", bold=True) for share_id in missing_ids
        )
        click.echo(
            f"""\n🚸 {click.style('The following secret shares will not be '
                                    'decrypted:', fg='yellow')} {fmt_ids}\n"""
        )

    return recover_object_decryption_key_from_encrypted_shares(
        manifest.decryption_key_shares,
        functools.partial(_share_decryption_keys_provider, yubikey_share_ids),
        decrypted_mnemonic_processor=decrypted_mnemonic_processor,
        known_mnemonics=known_mnemonics,
    )


def get_object_decryption_key_provider(ctx) -> ObjectDecryptionKeyProvider:
    import functools

    secrets = ctx.params.get("secret")
    identity_files = ctx.params.get("identity")
    object_decryption_key_provider: ObjectDecryptionKeyProvider = functools.partial(
        prompting_object_decryption_key_provider,
        known_mnemonics=secrets,
        identity_files=identity_files,
    )
    decryption_key = ctx.params.get("decryption_key")
    if decryption_key:
        if not decryption_key.lower().startswith("age-secret-key-"):
            ctx.fail(
                "The given decryption key does not look like a decryption key. "
                "It should start with “AGE-SECRET-KEY-”"
            )

        def known_key_provider(_):
            return decryption_key

        object_decryption_key_provider = known_key_provider
    return object_decryption_key_provider


class ContentSWHID(click.ParamType):
    name = "swhid of a content object"

    def convert(self, value, param, ctx):
        from swh.model.swhids import ExtendedObjectType, ExtendedSWHID, ValidationError

        try:
            swhid = ExtendedSWHID.from_string(value)
        except ValidationError:
            self.fail(f"expected SWHID, got {value!r}", param, ctx)
        if swhid.object_type != ExtendedObjectType.CONTENT:
            self.fail("We can only extract data for Content objects", param, ctx)
        return swhid


@recovery_bundle_cli_group.command(name="extract-content")
@click.option(
    "-o",
    "--output",
    type=click.File("wb"),
    metavar="FILE",
    required=True,
    help="write data to FILE",
)
@click.option(
    "--decryption-key",
    metavar="AGE_SECRET_KEY",
    help="use the given decryption key instead of the bundle shared secrets",
)
@click.option(
    "-s",
    "--secret",
    metavar="MNEMONIC",
    multiple=True,
    help="Known shared secret. May be repeated.",
)
@click.option(
    "-i",
    "--identity",
    metavar="IDENTITY",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    multiple=True,
    help="Path to file with age identities. May be repeated.",
)
@click.argument(
    "recovery-bundle",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=True,
)
@click.argument(
    "SWHID",
    type=ContentSWHID(),
    required=True,
)
@click.pass_context
def extract_content(
    ctx,
    output,
    recovery_bundle,
    swhid,
    decryption_key=None,
    identity=None,
    secret=None,
) -> None:
    """Extract data from content stored in a recovery bundle."""
    from .recovery_bundle import RecoveryBundle, WrongDecryptionKey

    secret_key_provider = get_object_decryption_key_provider(ctx)
    bundle = RecoveryBundle(recovery_bundle, secret_key_provider)

    if swhid not in bundle.swhids:
        click.secho(
            f"“{swhid}” is not in the recovery bundle", err=True, fg="red", bold=True
        )
        ctx.exit(1)

    try:
        bundle.write_content_data(swhid, output)
    except WrongDecryptionKey:
        click.secho(
            f"Wrong decryption key for this bundle ({bundle.removal_identifier})",
            err=True,
            fg="red",
            bold=True,
        )
        ctx.exit(2)


@recovery_bundle_cli_group.command(name="restore")
@click.option(
    "--decryption-key",
    metavar="AGE_SECRET_KEY",
    help="use the given decryption key instead of the bundle shared secrets",
)
@click.option(
    "-s",
    "--secret",
    metavar="MNEMONIC",
    multiple=True,
    help="Known shared secret. May be repeated.",
)
@click.option(
    "-i",
    "--identity",
    metavar="IDENTITY",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    multiple=True,
    help="Path to file with age identities. May be repeated.",
)
@click.argument(
    "recovery-bundle",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=True,
)
@click.pass_context
def restore(
    ctx, recovery_bundle, decryption_key=None, identity=None, secret=None
) -> None:
    """Restore a recovery bundle to Software Heritage archive."""
    from .recovery_bundle import (
        RecoveryBundle,
        UnsupportedFeatureException,
        WrongDecryptionKey,
    )

    conf = ctx.obj["config"]
    from swh.storage import get_storage

    restoration_storage = get_storage(**conf["restoration_storage"])

    secret_key_provider = get_object_decryption_key_provider(ctx)
    bundle = RecoveryBundle(recovery_bundle, secret_key_provider)
    try:
        missing = bundle.get_missing_referenced_objects(restoration_storage)
        if len(missing) > 0:
            click.secho(
                "Objects to be restored are referencing objects that "
                "are missing from storage:",
                fg="yellow",
                bold=True,
            )
            for swhid in missing:
                click.secho(f"- {swhid}", fg="yellow")
            click.confirm(
                click.style(
                    "Proceed with restoration though it will create "
                    "references to missing objects?",
                    fg="yellow",
                    bold=True,
                ),
                abort=True,
            )
    except UnsupportedFeatureException:
        click.secho(
            "Skipping checks for missing referenced objects: "
            f"recovery bundle “{recovery_bundle}” is too old.",
            fg="yellow",
            bold=True,
        )
    try:
        bundle.restore(restoration_storage, progressbar)
    except WrongDecryptionKey:
        click.echo(
            f"Wrong decryption key for this bundle ({bundle.removal_identifier})"
        )
        ctx.exit(2)


@recovery_bundle_cli_group.command(name="resume-removal")
@click.option(
    "--decryption-key",
    metavar="AGE_SECRET_KEY",
    prompt=True,
    help="use the given decryption key instead of the bundle shared secrets",
    envvar="SWH_BUNDLE_DECRYPTION_KEY",
)
@click.argument(
    "recovery-bundle",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=True,
)
@click.pass_context
def resume_removal(
    ctx,
    recovery_bundle,
    decryption_key=None,
) -> None:
    """Resume a removal operation from a recovery bundle."""
    from swh.journal.writer import JournalWriterInterface, get_journal_writer

    from .notifications import RemovalNotification
    from .recovery_bundle import WrongDecryptionKey

    remover = get_remover(ctx)
    journal_writer: Optional[JournalWriterInterface] = None
    if "journal_writer" in ctx.obj["config"]:
        cfg = ctx.obj["config"]["journal_writer"]
        journal_writer = get_journal_writer(**cfg)

    try:
        bundle = remover.register_objects_from_bundle(
            recovery_bundle_path=recovery_bundle, object_secret_key=decryption_key
        )
    except WrongDecryptionKey:
        click.echo("Wrong decryption key for this bundle")
        ctx.exit(2)
    try:
        remover.remove()
    except Exception as e:
        click.secho(str(e), err=True, fg="red", bold=True)
        remover.restore_recovery_bundle()
        ctx.exit(1)
    else:
        if journal_writer is not None:
            notif = RemovalNotification(
                removal_identifier=bundle.removal_identifier,
                reason=bundle.reason or "",
                requested=bundle.requested,
                removed_objects=remover.swhids_to_remove,
            )
            journal_writer.write_addition("removal_notification", notif)
            journal_writer.flush()


def _strip_rage_report(output):
    # rage prompts for report when it errors like this:
    #   [ Did rage not do what you expected? Could an error be more useful? ]
    #   [ Tell us: https://str4d.xyz/rage/report                            ]
    # This can be confusing in our case so strip them from the output.
    return b"\n".join(
        line
        for line in output.split(b"\n")
        if not line.startswith(b"[") and not line.endswith(b"]")
    )


@recovery_bundle_cli_group.command(name="recover-decryption-key")
@click.option(
    "-s",
    "--secret",
    metavar="MNEMONIC",
    multiple=True,
    help="Known shared secret. May be repeated.",
)
@click.option(
    "-i",
    "--identity",
    metavar="IDENTITY",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    multiple=True,
    help="Path to file with age identities. May be repeated.",
)
@click.option(
    "--show-recovered-secrets",
    is_flag=True,
    default=False,
    help="Show recovered shared secrets. Useful for remote/distributed recoveries.",
)
@click.argument(
    "recovery-bundle",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    required=True,
)
def recover_decryption_key(
    recovery_bundle, secret, identity, show_recovered_secrets
) -> None:
    """Recover the decryption key using shared secrets."""
    import subprocess
    import sys

    from .recovery_bundle import RecoveryBundle

    def object_decryption_key_provider(*args, **kwargs):
        kwargs["known_mnemonics"] = list(secret)
        kwargs["identity_files"] = list(identity)
        kwargs["show_decrypted_mnemonics"] = show_recovered_secrets
        return prompting_object_decryption_key_provider(*args, **kwargs)

    try:
        bundle = RecoveryBundle(recovery_bundle, object_decryption_key_provider)
        decryption_key = bundle.object_decryption_key
        click.echo(
            f"\n🔓 Recovered decryption key:\n{click.style(decryption_key, bold=True)}"
        )
    except subprocess.CalledProcessError as ex:
        if "rage" not in ex.cmd[0] and ex.cmd[1] != "--decrypt":
            raise
        click.echo(
            f"""💥 {click.style('rage decryption failed:', bold=True, fg='red')}"""
        )
        click.echo(_strip_rage_report(ex.stderr))
        sys.exit(1)


@recovery_bundle_cli_group.command(name="rollover")
@click.option(
    "--decryption-key",
    metavar="AGE_SECRET_KEY",
    help="use the given decryption key instead of the bundle shared secrets",
)
@click.option(
    "-s",
    "--secret",
    metavar="MNEMONIC",
    multiple=True,
    help="Known shared secret. May be repeated.",
)
@click.option(
    "-i",
    "--identity",
    metavar="IDENTITY",
    type=click.Path(exists=True, readable=True, dir_okay=False),
    multiple=True,
    help="Path to file with age identities. May be repeated.",
)
@click.argument(
    "recovery-bundles",
    metavar="[RECOVERY_BUNDLE]…",
    type=click.Path(exists=True, dir_okay=False, readable=True),
    nargs=-1,
)
@click.pass_context
def rollover(
    ctx, recovery_bundles, decryption_key=None, identity=None, secret=None
) -> None:
    """Rollover recovery bundles to new shared secrets."""
    conf = ctx.obj["config"]

    from .recovery_bundle import RecoveryBundle, SecretSharing, WrongDecryptionKey

    secret_key_provider = get_object_decryption_key_provider(ctx)
    secret_sharing = SecretSharing.from_dict(conf["recovery_bundles"]["secret_sharing"])
    click.secho("New shared secret holders:")
    for share_id in sorted(secret_sharing.share_ids):
        click.echo(f"- {click.style(share_id, fg='magenta', bold=True)}")
    click.confirm(
        click.style(
            "Proceed with rolling over the shared secrets?",
            fg="yellow",
            bold=True,
        ),
        abort=True,
    )
    for recovery_bundle in recovery_bundles:
        bundle = RecoveryBundle(recovery_bundle, secret_key_provider)
        # Ensure that we can decrypt at least some objects with the provided key
        try:
            origin = list(bundle.origins())
            assert len(origin) > 0, "Oops! No Origin objects in this recovery bundle."
        except WrongDecryptionKey:
            click.secho(
                f"Wrong decryption key for this bundle ({bundle.removal_identifier})",
                err=True,
                fg="red",
                bold=True,
            )
            ctx.exit(2)
        bundle.rollover(secret_sharing)
        click.secho("Shared secrets for ", fg="green", nl=False)
        click.secho(bundle.removal_identifier, fg="green", bold=True, nl=False)
        click.secho(" have been rolled over.", fg="green")


@alter_cli_group.group(
    name="handle-removal-notification", context_settings=CONTEXT_SETTINGS
)
@click.pass_context
def handle_removal_notification_cli_group(ctx):
    """Tools to handle removal notifications."""
    return ctx


@handle_removal_notification_cli_group.command(name="remove")
@click.option(
    "--recovery-bundle",
    type=click.Path(dir_okay=False),
    required=True,
)
@click.option(
    "--allow-empty-content-objects/--disallow-empty-content-objects",
    "allow_empty_content_objects",
    default=False,
    help="Create recovery bundle even when data for Content object cannot be found",
)
@click.option(
    "--ignore-requested",
    "ignore_requested",
    metavar="ORIGIN_OR_SWHID",
    multiple=True,
    type=SwhidOrUrlParamType(),
    help="object that should be ignored from the list of "
    "requested objects to be removed",
)
@click.option(
    "--ignore",
    "ignore_backends",
    default=None,
    multiple=True,
    type=click.Choice(["search", "journal", "graph"], case_sensitive=False),
    help=(
        "Do not make the given backend mandatory when checking the configuration; "
        "this command is usually meant to remove objects from all possible data silos, "
        "so the default behavior is to have all of them mandatory in the configuration "
        "file. Using this option allows to explicitly ignore the given 'backend'"
    ),
)
@click.option(
    "--recompute/--no-recompute",
    default=False,
    help=(
        "If set, recompute locally the list of swhids to remove to apply "
        "the given removal request, otherwise, use the list of swhids "
        "sent in the notification message. Note that an access to swh-graph "
        "is required to reocmpute this list. Implies '--ignore graph' if not set."
    ),
)
@click.argument(
    "removal-identifier",
    required=True,
)
@click.pass_context
def handle_removal_notification_with_removal(
    ctx,
    recovery_bundle,
    removal_identifier,
    allow_empty_content_objects,
    ignore_requested,
    ignore_backends,
    recompute,
):
    """Handle removal notification by removing the request objects."""

    from .inventory import RootsNotFound, StuckInventoryException
    from .operations import MaskingRequestNotFound, RemoverError
    from .recovery_bundle import ContentDataNotFound, SecretSharing

    ignore_backends = set(ignore_backends or [])

    if recompute:
        ignore_backends.discard("graph")
    else:
        ignore_backends.add("graph")

    remover = get_remover(
        ctx,
        require_masking_admin=True,
        ignore_backends=ignore_backends,
    )

    try:
        secret_sharing = SecretSharing.from_dict(
            ctx.obj["config"]["recovery_bundles"]["secret_sharing"]
        )
    except ValueError as e:
        raise click.ClickException(f"Wrong secret sharing configuration: {e.args[0]}")

    try:
        p = pathlib.Path(recovery_bundle)
        p.touch(exist_ok=False)
        p.unlink()
    except FileExistsError:
        raise click.ClickException(f"File “{recovery_bundle}” already exists")
    except PermissionError:
        raise click.ClickException(f"Permission denied: “{recovery_bundle}”")

    try:
        remover.handle_removal_notification_with_removal(
            notification_removal_identifier=removal_identifier,
            secret_sharing=secret_sharing,
            recovery_bundle_path=recovery_bundle,
            ignore_requested=ignore_requested or [],
            allow_empty_content_objects=allow_empty_content_objects,
            recompute_swhids_to_remove=recompute,
        )
    except MaskingRequestNotFound as e:
        click.secho(
            f"Masking request “{e.masking_request_slug}” has not been found.",
            err=True,
            fg="red",
        )
        click.secho(
            f"Please double-check that “{removal_identifier}” is the identifier of "
            "the notification. Otherwise, there probably was an error with the "
            "processing of the notification."
        )
        ctx.exit(1)
    except RootsNotFound:
        # handle_removal_notification_with_removal() has already
        # displayed the missing objects
        click.secho(
            "You might want to use “--ignore-requested=https://…” "
            "to specify which requested object should be ignored.",
            err=True,
            fg="yellow",
            bold="True",
        )
        ctx.exit(1)
    except StuckInventoryException as e:
        click.secho(
            "Inventory phase got stuck. "
            "Unable to learn the complete set of what these objects reference:",
            err=True,
            fg="red",
            bold=True,
        )
        click.secho("\n".join(f"- {swhid}" for swhid in e.swhids), err=True, fg="red")
        ctx.exit(1)
    except ContentDataNotFound as e:
        click.secho(
            f"Content “{e.swhid}” exists, but its data was not found.",
            err=True,
            fg="red",
            bold=True,
        )
        click.secho(
            "Consider using `--allow-empty-content-objects` but only "
            "if the above is expected.",
            err=True,
            fg="yellow",
        )
        ctx.exit(1)
    except RemoverError:
        # Showing the exception and rolling back has been handled by
        # handle_removal_notification_with_removal()
        ctx.exit(1)
    except Exception as e:
        raise e


@handle_removal_notification_cli_group.command(name="restrict-permanently")
@click.argument(
    "removal-identifier",
    required=True,
)
@click.pass_context
def handle_removal_notification_with_permanent_restriction(
    ctx,
    removal_identifier,
):
    """Handle removal notification by permanently restricting access
    to objects removed from the main archive."""

    from swh.storage.proxies.masking.db import MaskedState

    from .operations import MaskingRequestNotFound

    remover = get_remover(
        ctx, require_masking_admin=True, ignore_backends=["graph", "search", "journal"]
    )

    try:
        remover.handle_removal_notification_by_changing_masked_status(
            notification_removal_identifier=removal_identifier,
            masked_state=MaskedState.RESTRICTED,
        )
    except MaskingRequestNotFound as e:
        click.secho(
            f"Masking request “{e.masking_request_slug}” has not been found.",
            err=True,
            fg="red",
        )
        click.secho(
            f"Please double-check that “{removal_identifier}” is the identifier of "
            "the notification. Otherwise, there probably was an error with the "
            "processing of the notification."
        )
        ctx.exit(1)
    click.secho(
        f"Removal notification “{removal_identifier}” handled by "
        "permanently restricting access to the objects removed from the main archive.",
        fg="green",
        bold=True,
    )


@handle_removal_notification_cli_group.command(name="dismiss")
@click.argument(
    "removal-identifier",
    required=True,
)
@click.pass_context
def handle_removal_notification_with_dismissal(
    ctx,
    removal_identifier,
):
    """Handle removal notification by making the objects removed from the
    main archive visible again."""

    from swh.storage.proxies.masking.db import MaskedState

    from .operations import MaskingRequestNotFound

    remover = get_remover(
        ctx, require_masking_admin=True, ignore_backends=["graph", "search", "journal"]
    )

    try:
        remover.handle_removal_notification_by_changing_masked_status(
            notification_removal_identifier=removal_identifier,
            masked_state=MaskedState.VISIBLE,
        )
    except MaskingRequestNotFound as e:
        click.secho(
            f"Masking request “{e.masking_request_slug}” has not been found.",
            err=True,
            fg="red",
        )
        click.secho(
            f"Please double-check that “{removal_identifier}” is the identifier of "
            "the notification. Otherwise, there probably was an error with the "
            "processing of the notification."
        )
        ctx.exit(1)
    click.secho(
        f"Removal notification “{removal_identifier}” has been dismissed: "
        "objects removed from the main archive are now visible again.",
        fg="green",
        bold=True,
    )
